
from pcbnew import *
import random
import pprint


class testundoredo0(ActionPlugin):

    def defaults(self):
        self.name = "Test Undo/Redo: Remove footprints"
        self.category = "Test Undo/Redo"
        self.description = ""

    def Run(self):
        pcb = GetBoard()

        for module in pcb.GetFootprints():
            pcb.RemoveNative(module)


class testundoredo1(ActionPlugin):

    def defaults(self):
        self.name = "Test Undo/Redo: Remove all board items"
        self.category = "Test Undo/Redo"
        self.description = ""

    def Run(self):
        pcb = GetBoard()

        while pcb.GetAreaCount() > 0:
            area = pcb.GetArea(0)
            pcb.RemoveNative(area)

        for module in pcb.GetFootprints():
            pcb.RemoveNative(module)

        for track in pcb.GetTracks():
            pcb.RemoveNative(track)

        for draw in pcb.GetDrawings():
            pcb.RemoveNative(draw)


class testundoredo2(ActionPlugin):

    def defaults(self):
        self.name = "Test Undo/Redo: Generate random content"
        self.category = "Test Undo/Redo"
        self.description = ""

    def createFPCXFootprint(self,pads):
        size_025_160mm = VECTOR2I_MM(0.25,1.6)
        size_150_200mm = VECTOR2I_MM(1.50,2.0)
        # create a new footprint, its parent is our previously created pcb
        footprint = FOOTPRINT(self.pcb)
        footprint.SetReference("FPC"+str(pads))   # give it a reference name
        footprint.Reference().SetPosition(VECTOR2I_MM(-1,-1))
        self.pcb.Add(footprint)             # add it to our pcb
        m_pos = VECTOR2I_MM(0,0)        #random.randint(10,200),random.randint(10,200))
        footprint.SetPosition(m_pos)

        # create a pad array and add it to the footprint


        def smdRectPad(footprint,size,pos,name):
            pad = PAD(footprint)
            pad.SetSize(size)
            pad.SetShape(PAD_SHAPE_RECT)
            pad.SetAttribute(PAD_ATTRIB_SMD)
            pad.SetLayerSet(pad.SMDMask())
            pad.SetPosition(pos)
            pad.SetPadName(name)
            return pad

        for n in range (0,pads):
            pad = smdRectPad( footprint,size_025_160mm, VECTOR2I_MM(0.5*n,0), str(n+1) )
            footprint.Add(pad)


        pad_s0 = smdRectPad(footprint,size_150_200mm,VECTOR2I_MM(-1.6,1.3), "0")
        pad_s1 = smdRectPad(footprint,size_150_200mm,VECTOR2I_MM((pads-1)*0.5+1.6,1.3), "0")
        footprint.Add(pad_s0)
        footprint.Add(pad_s1)

        e = PCB_SHAPE(footprint)
        e.SetStart(VECTOR2I_MM(-1,0))
        e.SetEnd(VECTOR2I_MM(0,0))
        e.SetWidth(FromMM(0.2))
        e.SetLayer(F_SilkS)
        e.SetShape(S_SEGMENT)
        footprint.Add(e)
        footprint.SetPosition(VECTOR2I_MM(random.randint(20,200),random.randint(20,150)))
        return footprint



    def Run(self):
        self.pcb = GetBoard()
        random.seed()

        for i in range(10):
            seg = PCB_SHAPE()
            seg.SetLayer( random.choice([Edge_Cuts,Cmts_User,Eco1_User,Eco2_User]) )
            seg.SetStart( VECTOR2I_MM( random.randint(10,100),
                                random.randint(10,100) ) )
            seg.SetEnd( VECTOR2I_MM( random.randint(10,100),
                                random.randint(10,100) ) )
            self.pcb.Add( seg )

            if i%2 == 0:
                t = PCB_TRACK(None)
            else:
                t = PCB_VIA(None)
                #t.SetLayerPair(segments['layerPair'][0],segments['layerPair'][1])
                t.SetViaType(VIATYPE_THROUGH)
                t.SetDrill(FromMM(random.randint(1,20)/10.0))

            t.SetStart(VECTOR2I_MM(random.randint(100,150),random.randint(100,150)))
            t.SetEnd(VECTOR2I_MM(random.randint(100,150),random.randint(100,150)))
            t.SetWidth(FromMM(random.randint(1,15)/10.0))
            t.SetLayer(random.choice([F_Cu,B_Cu]))
            self.pcb.Add(t)

            self.createFPCXFootprint(random.randint(2,40))


class testundoredo3(ActionPlugin):

    def defaults(self):
        self.name = "Test Undo/Redo: Move elements randomly"
        self.category = "Test Undo/Redo"
        self.description = ""

    def Run(self):
        pcb = GetBoard()

        for i in range(0,pcb.GetAreaCount()):
            area = pcb.GetArea(i)
            area.Move(VECTOR2I_MM(random.randint(-20,20),random.randint(-20,20)))

        for footprint in pcb.GetFootprints():
            footprint.Move(VECTOR2I_MM(random.randint(-20,20),random.randint(-20,20)))
            if random.randint(0,10) > 5:
                footprint.Flip(footprint.GetPosition(), True)

        for track in pcb.GetTracks():
            track.Move(VECTOR2I_MM(random.randint(-20,20),random.randint(-20,20)))

        for draw in pcb.GetDrawings():
            draw.Move(VECTOR2I_MM(random.randint(-20,20),random.randint(-20,20)))



testundoredo0().register()
testundoredo1().register()
testundoredo2().register()
testundoredo3().register()



from __future__ import annotations

import pytest

import env  # noqa: F401
from pybind11_tests import class_sh_disowning_mi as m


def test_diamond_inheritance():
    # Very similar to test_multiple_inheritance.py:test_diamond_inheritance.
    d = m.D()
    assert d is d.d()
    assert d is d.c0()
    assert d is d.c1()
    assert d is d.b()
    assert d is d.c0().b()
    assert d is d.c1().b()
    assert d is d.c0().c1().b().c0().b()


def is_disowned(callable_method):
    try:
        callable_method()
    except ValueError as e:
        assert "Python instance was disowned" in str(e)  # noqa: PT017
        return True
    return False


def test_disown_b():
    b = m.B()
    assert b.get() == 10
    m.disown_b(b)
    assert is_disowned(b.get)


@pytest.mark.parametrize("var_to_disown", ["c0", "b"])
def test_disown_c0(var_to_disown):
    c0 = m.C0()
    assert c0.get() == 1020
    b = c0.b()
    m.disown_b(locals()[var_to_disown])
    assert is_disowned(c0.get)
    assert is_disowned(b.get)


@pytest.mark.parametrize("var_to_disown", ["c1", "b"])
def test_disown_c1(var_to_disown):
    c1 = m.C1()
    assert c1.get() == 1021
    b = c1.b()
    m.disown_b(locals()[var_to_disown])
    assert is_disowned(c1.get)
    assert is_disowned(b.get)


@pytest.mark.parametrize("var_to_disown", ["d", "c1", "c0", "b"])
def test_disown_d(var_to_disown):
    d = m.D()
    assert d.get() == 10202130
    b = d.b()
    c0 = d.c0()
    c1 = d.c1()
    m.disown_b(locals()[var_to_disown])
    assert is_disowned(d.get)
    assert is_disowned(c1.get)
    assert is_disowned(c0.get)
    assert is_disowned(b.get)


# Based on test_multiple_inheritance.py:test_multiple_inheritance_python.
class MI1(m.Base1, m.Base2):
    def __init__(self, i, j):
        m.Base1.__init__(self, i)
        m.Base2.__init__(self, j)


class B1:
    def v(self):
        return 1


class MI2(B1, m.Base1, m.Base2):
    def __init__(self, i, j):
        B1.__init__(self)
        m.Base1.__init__(self, i)
        m.Base2.__init__(self, j)


class MI3(MI2):
    def __init__(self, i, j):
        MI2.__init__(self, i, j)


class MI4(MI3, m.Base2):
    def __init__(self, i, j):
        MI3.__init__(self, i, j)
        # This should be ignored (Base2 is already initialized via MI2):
        m.Base2.__init__(self, i + 100)


class MI5(m.Base2, B1, m.Base1):
    def __init__(self, i, j):
        B1.__init__(self)
        m.Base1.__init__(self, i)
        m.Base2.__init__(self, j)


class MI6(m.Base2, B1):
    def __init__(self, i):
        m.Base2.__init__(self, i)
        B1.__init__(self)


class B2(B1):
    def v(self):
        return 2


class B3:
    def v(self):
        return 3


class B4(B3, B2):
    def v(self):
        return 4


class MI7(B4, MI6):
    def __init__(self, i):
        B4.__init__(self)
        MI6.__init__(self, i)


class MI8(MI6, B3):
    def __init__(self, i):
        MI6.__init__(self, i)
        B3.__init__(self)


class MI8b(B3, MI6):
    def __init__(self, i):
        B3.__init__(self)
        MI6.__init__(self, i)


@pytest.mark.xfail("env.PYPY")
def test_multiple_inheritance_python():
    # Based on test_multiple_inheritance.py:test_multiple_inheritance_python.
    # Exercises values_and_holders with 2 value_and_holder instances.

    mi1 = MI1(1, 2)
    assert mi1.foo() == 1
    assert mi1.bar() == 2

    mi2 = MI2(3, 4)
    assert mi2.v() == 1
    assert mi2.foo() == 3
    assert mi2.bar() == 4

    mi3 = MI3(5, 6)
    assert mi3.v() == 1
    assert mi3.foo() == 5
    assert mi3.bar() == 6

    mi4 = MI4(7, 8)
    assert mi4.v() == 1
    assert mi4.foo() == 7
    assert mi4.bar() == 8

    mi5 = MI5(10, 11)
    assert mi5.v() == 1
    assert mi5.foo() == 10
    assert mi5.bar() == 11

    mi6 = MI6(12)
    assert mi6.v() == 1
    assert mi6.bar() == 12

    mi7 = MI7(13)
    assert mi7.v() == 4
    assert mi7.bar() == 13

    mi8 = MI8(14)
    assert mi8.v() == 1
    assert mi8.bar() == 14

    mi8b = MI8b(15)
    assert mi8b.v() == 3
    assert mi8b.bar() == 15


DISOWN_CLS_I_J_V_LIST = [
    (MI1, 1, 2, None),
    (MI2, 3, 4, 1),
    (MI3, 5, 6, 1),
    (MI4, 7, 8, 1),
    (MI5, 10, 11, 1),
]


@pytest.mark.xfail("env.PYPY", strict=False)
@pytest.mark.parametrize(("cls", "i", "j", "v"), DISOWN_CLS_I_J_V_LIST)
def test_disown_base1_first(cls, i, j, v):
    obj = cls(i, j)
    assert obj.foo() == i
    assert m.disown_base1(obj) == 2000 * i + 1
    assert is_disowned(obj.foo)
    assert obj.bar() == j
    assert m.disown_base2(obj) == 2000 * j + 2
    assert is_disowned(obj.bar)
    if v is not None:
        assert obj.v() == v


@pytest.mark.xfail("env.PYPY", strict=False)
@pytest.mark.parametrize(("cls", "i", "j", "v"), DISOWN_CLS_I_J_V_LIST)
def test_disown_base2_first(cls, i, j, v):
    obj = cls(i, j)
    assert obj.bar() == j
    assert m.disown_base2(obj) == 2000 * j + 2
    assert is_disowned(obj.bar)
    assert obj.foo() == i
    assert m.disown_base1(obj) == 2000 * i + 1
    assert is_disowned(obj.foo)
    if v is not None:
        assert obj.v() == v


@pytest.mark.xfail("env.PYPY", strict=False)
@pytest.mark.parametrize(
    ("cls", "j", "v"),
    [
        (MI6, 12, 1),
        (MI7, 13, 4),
        (MI8, 14, 1),
        (MI8b, 15, 3),
    ],
)
def test_disown_base2(cls, j, v):
    obj = cls(j)
    assert obj.bar() == j
    assert m.disown_base2(obj) == 2000 * j + 2
    assert is_disowned(obj.bar)
    assert obj.v() == v



#
# Example python script to generate an equivalent XML document from XML input
#
# Example: Round value robin, XML to XML conversion with partial value monging
#

from __future__ import print_function

# Import the KiCad python helper module and the csv formatter
import kicad_netlist_reader
import sys

def checkvalue(self):
    """Check values, and replace with preferred/consistent values"""
    ref = self.getRef()
    r = ref.split("R")
    c = ref.split("C")
    v = self.getValue()

    # Common to all values - convert decimation if necessary
    dec = v.split(",")
    if (len(dec) == 2):
        newval = dec[0] + "." + dec[1]
        self.setValue(newval)
        v = self.getValue()

    if len(r) == 2 and r[1].isdigit():
        # This is a resistor - make values consistent
        # If the value is a pure value, add R to the end of the value
        if v.isdigit():
            i = int(v)
            if (i > 1000000):
                i = i / 1000000
                v = str(i) + "M"
            if (i > 1000):
                i = i / 1000
                v = str(i) + "K"
            else:
                v = str(i) + "R"

            self.setValue(v)
        else:
            # Get the multiplier character
            multiplier = v[len(v) - 1]
            v = v.strip(multiplier)
            v = v.split(".")
            if (len(v) == 2):
                newval = v[0] + multiplier + v[1]
                self.setValue(newval)
                v = self.getValue()



# Give components a new method for checking the values (this could easily be a
# Company Part Number generator method instead)
kicad_netlist_reader.comp.checkvalue = checkvalue

# Generate an instance of a generic netlist, and load the netlist tree from
# the command line option. If the file doesn't exist, execution will stop
net = kicad_netlist_reader.netlist(sys.argv[1])

# Open a file to write to, if the file cannot be opened output to stdout
# instead
canOpenFile = True
try:
    f = open(sys.argv[2], 'w')
except IOError:
    e = "Can't open output file for writing: " + sys.argv[2]
    print(__file__, ":", e, file=sys.stderr)
    f = sys.stdout
    canOpenFile = False

for c in net.components:
    c.checkvalue()

print(net.formatXML(), file=f)

if not canOpenFile:
    sys.exit(1)


"""pytest configuration

Extends output capture as needed by pybind11: ignore constructors, optional unordered lines.
Adds docstring and exceptions message sanitizers.
"""

from __future__ import annotations

import contextlib
import difflib
import gc
import importlib.metadata
import multiprocessing
import re
import sys
import sysconfig
import textwrap
import traceback
from typing import Callable

import pytest

# Early diagnostic for failed imports
try:
    import pybind11_tests
except Exception:
    # pytest does not show the traceback without this.
    traceback.print_exc()
    raise


@pytest.fixture(scope="session", autouse=True)
def use_multiprocessing_forkserver_on_linux():
    if sys.platform != "linux" or sys.implementation.name == "graalpy":
        # The default on Windows, macOS and GraalPy is "spawn": If it's not broken, don't fix it.
        return

    # Full background: https://github.com/pybind/pybind11/issues/4105#issuecomment-1301004592
    # In a nutshell: fork() after starting threads == flakiness in the form of deadlocks.
    # It is actually a well-known pitfall, unfortunately without guard rails.
    # "forkserver" is more performant than "spawn" (~9s vs ~13s for tests/test_gil_scoped.py,
    # visit the issuecomment link above for details).
    multiprocessing.set_start_method("forkserver")


_long_marker = re.compile(r"([0-9])L")
_hexadecimal = re.compile(r"0x[0-9a-fA-F]+")

# Avoid collecting Python3 only files
collect_ignore = []


def _strip_and_dedent(s):
    """For triple-quote strings"""
    return textwrap.dedent(s.lstrip("\n").rstrip())


def _split_and_sort(s):
    """For output which does not require specific line order"""
    return sorted(_strip_and_dedent(s).splitlines())


def _make_explanation(a, b):
    """Explanation for a failed assert -- the a and b arguments are List[str]"""
    return ["--- actual / +++ expected"] + [
        line.strip("\n") for line in difflib.ndiff(a, b)
    ]


class Output:
    """Basic output post-processing and comparison"""

    def __init__(self, string):
        self.string = string
        self.explanation = []

    def __str__(self):
        return self.string

    __hash__ = None

    def __eq__(self, other):
        # Ignore constructor/destructor output which is prefixed with "###"
        a = [
            line
            for line in self.string.strip().splitlines()
            if not line.startswith("###")
        ]
        b = _strip_and_dedent(other).splitlines()
        if a == b:
            return True
        self.explanation = _make_explanation(a, b)
        return False


class Unordered(Output):
    """Custom comparison for output without strict line ordering"""

    __hash__ = None

    def __eq__(self, other):
        a = _split_and_sort(self.string)
        b = _split_and_sort(other)
        if a == b:
            return True
        self.explanation = _make_explanation(a, b)
        return False


class Capture:
    def __init__(self, capfd):
        self.capfd = capfd
        self.out = ""
        self.err = ""

    def __enter__(self):
        self.capfd.readouterr()
        return self

    def __exit__(self, *args):
        self.out, self.err = self.capfd.readouterr()

    __hash__ = None

    def __eq__(self, other):
        a = Output(self.out)
        b = other
        if a == b:
            return True
        self.explanation = a.explanation
        return False

    def __str__(self):
        return self.out

    def __contains__(self, item):
        return item in self.out

    @property
    def unordered(self):
        return Unordered(self.out)

    @property
    def stderr(self):
        return Output(self.err)


@pytest.fixture
def capture(capsys):
    """Extended `capsys` with context manager and custom equality operators"""
    return Capture(capsys)


class SanitizedString:
    def __init__(self, sanitizer):
        self.sanitizer = sanitizer
        self.string = ""
        self.explanation = []

    def __call__(self, thing):
        self.string = self.sanitizer(thing)
        return self

    __hash__ = None

    def __eq__(self, other):
        a = self.string
        b = _strip_and_dedent(other)
        if a == b:
            return True
        self.explanation = _make_explanation(a.splitlines(), b.splitlines())
        return False


def _sanitize_general(s):
    s = s.strip()
    s = s.replace("pybind11_tests.", "m.")
    return _long_marker.sub(r"\1", s)


def _sanitize_docstring(thing):
    s = thing.__doc__
    return _sanitize_general(s)


@pytest.fixture
def doc():
    """Sanitize docstrings and add custom failure explanation"""
    return SanitizedString(_sanitize_docstring)


def _sanitize_message(thing):
    s = str(thing)
    s = _sanitize_general(s)
    return _hexadecimal.sub("0", s)


@pytest.fixture
def msg():
    """Sanitize messages and add custom failure explanation"""
    return SanitizedString(_sanitize_message)


def pytest_assertrepr_compare(op, left, right):  # noqa: ARG001
    """Hook to insert custom failure explanation"""
    if hasattr(left, "explanation"):
        return left.explanation
    return None


def gc_collect():
    """Run the garbage collector three times (needed when running
    reference counting tests with PyPy)"""
    gc.collect()
    gc.collect()
    gc.collect()
    gc.collect()
    gc.collect()


def pytest_configure():
    pytest.suppress = contextlib.suppress
    pytest.gc_collect = gc_collect


def pytest_report_header():
    assert pybind11_tests.compiler_info is not None, (
        "Please update pybind11_tests.cpp if this assert fails."
    )
    interesting_packages = ("pybind11", "numpy", "scipy", "build")
    valid = []
    for package in sorted(interesting_packages):
        with contextlib.suppress(ModuleNotFoundError):
            valid.append(f"{package}=={importlib.metadata.version(package)}")
    reqs = " ".join(valid)

    cpp_info = [
        "C++ Info:",
        f"{pybind11_tests.compiler_info}",
        f"{pybind11_tests.cpp_std}",
        f"{pybind11_tests.PYBIND11_INTERNALS_ID}",
        f"PYBIND11_SIMPLE_GIL_MANAGEMENT={pybind11_tests.PYBIND11_SIMPLE_GIL_MANAGEMENT}",
    ]
    if "__graalpython__" in sys.modules:
        cpp_info.append(
            f"GraalPy version: {sys.modules['__graalpython__'].get_graalvm_version()}"
        )
    lines = [
        f"installed packages of interest: {reqs}",
        " ".join(cpp_info),
    ]
    if sysconfig.get_config_var("Py_GIL_DISABLED"):
        lines.append("free-threaded Python build")

    return lines


@pytest.fixture
def backport_typehints() -> Callable[[SanitizedString], SanitizedString]:
    d = {}
    if sys.version_info < (3, 13):
        d["typing_extensions.TypeIs"] = "typing.TypeIs"
        d["typing_extensions.CapsuleType"] = "types.CapsuleType"
    if sys.version_info < (3, 12):
        d["typing_extensions.Buffer"] = "collections.abc.Buffer"
    if sys.version_info < (3, 11):
        d["typing_extensions.Never"] = "typing.Never"
    if sys.version_info < (3, 10):
        d["typing_extensions.TypeGuard"] = "typing.TypeGuard"

    def backport(sanatized_string: SanitizedString) -> SanitizedString:
        for old, new in d.items():
            sanatized_string.string = sanatized_string.string.replace(old, new)

        return sanatized_string

    return backport



#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#

from __future__ import division
import pcbnew

import FootprintWizardBase as FPWbase
import PadArray as PA

class PadStaggeredZGridArray(PA.PadArray):
    """ Creates a staggered array of Pads with Z pad naming

                | pad pitch
                |   |
                o   o   o   o     ----
    staggered --> o   o   o   o   -- line_pitch
    line        o   o   o   o
                  o   o   o   o
    """
    def __init__(self, aPad, aPadCount, aLineCount, aLinePitch,
                 aPadPitch, aStagger=0, aCentre=pcbnew.VECTOR2I(0, 0)):
        """
        @param aPad         Template for all pads
        @param aPadCount    Overall pad count
        @param aLineCount   Number of lines
        @param aLinePitch   distance between lines
        @param aPadPitch    distance between pads
        @param aStagger     X stagger value for all odd lines
        @param aCentre      Center position

        """
        super(PadStaggeredZGridArray, self).__init__(aPad)

        self.padCount = int(aPadCount)
        self.lineCount = int(aLineCount)
        self.linePitch = aLinePitch
        self.padPitch = aPadPitch
        self.stagger = aStagger
        self.centre = aCentre


    # right to left, top to bottom
    def NamingFunction(self, aPadPos):
        return self.firstPadNum + aPadPos

    #relocate the pad and add it as many times as we need
    def AddPadsToModule(self, dc):
        pin1posX = self.centre.x - ((self.padPitch * (self.padCount // 2 - 1)) + self.stagger) / 2
        pin1posY = self.centre.y - self.linePitch * (self.lineCount - 1) / 2
        line = 0

        for padnum in range(0, self.padCount):
            if (line % 2) == 0:
                posX = pin1posX + ((padnum // 2) * self.padPitch)
            else:
                posX = pin1posX + self.stagger + ((padnum // 2) * self.padPitch)

            posY = pin1posY + (self.linePitch * line)

            pos = dc.TransformPoint(posX, posY)
            pad = self.GetPad(padnum == 0, pos)
            pad.SetName(self.GetName(padnum))
            self.AddPad(pad)

            line += 1

            if line >= self.lineCount:
                line = 0


class MicroMaTchWizard(FPWbase.FootprintWizard):
    padCountKey           = 'pad count'
    rowSpacingKey         = 'row spacing'
    padLengthKey          = 'pad length'
    padWidthKey           = 'pad width'
    padPitchKey           = 'pad pitch'
    staggerOffsetKey      = 'stagger_offset'
    withLockKey           = 'draw lock'

    def GetName(self):
        return "Micromatch SMD connectors"

    def GetDescription(self):
        return "Micromatch (with lock and w/o lock), footprint wizard"

    def GenerateParameterList(self):
        # defaults for a Micromatch package
        self.AddParam("Body", self.withLockKey,       self.uBool, True)
        self.AddParam("Pads", self.padCountKey,       self.uInteger, 8, multiple=2)

        #and override some of them
        self.AddParam("Pads", self.padWidthKey,       self.uMM, 1.5)
        self.AddParam("Pads", self.padLengthKey,      self.uMM, 3.0)
        self.AddParam("Pads", self.padPitchKey,       self.uMM, 2.54)
        self.AddParam("Pads", self.rowSpacingKey,     self.uMM, 5.2)
        self.AddParam("Pads", self.staggerOffsetKey,  self.uMM, 1.27)

    def CheckParameters(self):
        pass    # All checks are already taken care of!

    def GetValue(self):
        pad_count = self.parameters["Pads"][self.padCountKey]
        return "%s-%d" % ("ConnectorMicromatch", pad_count)

    def GetPad(self):
        padLength = self.parameters["Pads"][self.padLengthKey]
        padWidth  = self.parameters["Pads"][self.padWidthKey]
        return PA.PadMaker(self.module).SMDPad(
            padLength, padWidth, shape=pcbnew.PAD_SHAPE_RECT)

    def BuildThisFootprint(self):
        pads = self.parameters["Pads"]
        body = self.parameters["Body"]
        numPads = pads[self.padCountKey]
        padLength = pads[self.padLengthKey]
        rowPitch = pads[self.rowSpacingKey]
        padPitch = pads[self.padPitchKey]
        staggerOffset = pads[self.staggerOffsetKey]

        drawWithLock = body[self.withLockKey]
        numRows = 2

        # Use value to fill the modules description
        desc = self.GetValue()
        self.module.SetLibDescription(desc)
        self.module.SetAttributes( pcbnew.FP_SMD )

        # add in the pads
        pad = self.GetPad()

        array = PadStaggeredZGridArray(pad, numPads, numRows, rowPitch, padPitch, staggerOffset)
        array.AddPadsToModule(self.draw)

        # Draw connector outlineChassis
        width = pcbnew.FromMM(1.92) + (numPads * padPitch) / 2
        height = pcbnew.FromMM(5)

        self.draw.SetLineThickness( pcbnew.FromMM( 0.12 ) ) #Default per KLC F5.1 as of 12/2018

        # Left part
        #  --
        #  |
        #  ----
        self.draw.Polyline([(-width/2 + pcbnew.FromMM(0.5), -height/2),
                            (-width/2, -height/2),
                            (-width/2, height/2),
                            (-width/2 + pcbnew.FromMM(0.5) + padPitch / 2, height/2)])

        if drawWithLock :
            # Right part with pol slot
            #  ----
            #     [
            #    --
            self.draw.Polyline([(width/2 - pcbnew.FromMM(0.5) - padPitch / 2, -height/2),
                                (width/2,                                     -height/2),
                                (width/2,                                     -height/2 + pcbnew.FromMM(1.25)),
                                (width/2 - pcbnew.FromMM(0.7),                -height/2 + pcbnew.FromMM(1.25)),
                                (width/2 - pcbnew.FromMM(0.7),                 height/2 - pcbnew.FromMM(1.25)),
                                (width/2,                                      height/2 - pcbnew.FromMM(1.25)),
                                (width/2,                                      height/2),
                                (width/2 - pcbnew.FromMM(0.5),                 height/2)])
        else:
            # Right part without pol slot
            #  ----
            #     |
            #    --
            self.draw.Polyline([(width/2 - pcbnew.FromMM(0.5) - padPitch / 2, -height/2),
                                (width/2,                                     -height/2),
                                (width/2,                                      height/2),
                                (width/2 - pcbnew.FromMM(0.5),                 height/2)])

        # Courtyard
        self.draw.SetLayer(pcbnew.F_CrtYd)
        self.draw.SetLineThickness(pcbnew.FromMM(0.05))
        boxW = width                + pcbnew.FromMM(0.5)
        boxH = rowPitch + padLength + pcbnew.FromMM(0.5)

        # Round courtyard positions to 0.1 mm, rectangle will thus land on a 0.05mm grid
        pcbnew.PutOnGridMM(boxW, pcbnew.FromMM(0.10))
        pcbnew.PutOnGridMM(boxH, pcbnew.FromMM(0.10))
        self.draw.Box(0,0, boxW, boxH)

        #reference and value
        text_size = pcbnew.FromMM(1.0)  # According KLC
        text_offset = boxH/2 + text_size

        self.draw.Value(0, text_offset, text_size)
        self.draw.Reference(0, -text_offset, text_size)

        # Add a extra text (${REFERENCE}) on the F_Fab layer
        extra_text = pcbnew.PCB_TEXT( self.module )
        extra_text.SetLayer( pcbnew.F_Fab )
        extra_text.SetPosition( pcbnew.VECTOR2I( 0, 0) )
        extra_text.SetTextSize( pcbnew.VECTOR2I( text_size, text_size ) )
        extra_text.SetText( "${REFERENCE}" )
        self.module.Add( extra_text )


MicroMaTchWizard().register()



#!/usr/bin/env python

# the purpose of this script is rewriting the swig_import_helper
# call added by Swig version < 4.0 in pcbnew.py so it will not load
# _xxxxx.so/dso from inside a kicad executable
# because the kicad executable itself sill provide an _xxxxx module
# that's linked inside itself.
#
# for the normal module import it should work the same way with this
# fix in the swig_import_helper
#
# For Swig version >= 4.0 this method does not exist so nothing useful is made

from __future__ import print_function

from sys import argv,exit


if len(argv)<2:
    print("usage:")
    print("   fix_swig_imports.py file.py")
    print("")
    print("   will fix the swig import code for working inside KiCad")
    print("   where it happened that the external _pcbnew.so/dll was")
    print("   loaded too -and the internal _pcbnew module was to be used")
    exit(1)


filename = argv[1]

f = open(filename,"rb")
lines = f.readlines()
f.close()


doneOk = False

if (len(lines)<4000):
    print("still building")
    exit(0)

txt = b""

for l in lines:
    if l.startswith(b"if _swig_python_version_info < (2, 7, 0):"):     # ok with swig version >= 4.0.0
        l = l.replace(b"_swig_python_version_info < (2, 7, 0)", b"False")
        doneOk = True
    if l.startswith(b"if _swig_python_version_info >= (2, 7, 0):"):     # ok with swig version >= 3.0.10
        l = l.replace(b"_swig_python_version_info >= (2, 7, 0)", b"False")
        doneOk = True
    elif l.startswith(b"elif _swig_python_version_info >= (2, 6, 0):"):  # needed with swig version >= 3.0.10
        l = l.replace(b"_swig_python_version_info >= (2, 6, 0)", b"False")
        doneOk = True
    if l.startswith(b"if version_info >= (2, 7, 0):"):     # ok with swig version >= 3.0.9
        l = l.replace(b"version_info >= (2, 7, 0)", b"False")
        doneOk = True
    elif l.startswith(b"elif version_info >= (2, 6, 0):"):  # needed with swig version >= 3.0.9
        l = l.replace(b"version_info >= (2, 6, 0)", b"False")
        doneOk = True
    elif l.startswith(b"if version_info >= (2,6,0):"):     # ok with swig version <= 3.0.2
        l = l.replace(b"version_info >= (2,6,0)", b"False")
        doneOk = True
    elif l.startswith(b"if version_info >= (2, 6, 0):"): # needed with swig version 3.0.3
        l = l.replace(b"version_info >= (2, 6, 0)", b"False")
        doneOk = True
    elif l.startswith(b"if False:"):  # it was already patched?
        doneOk = True
    txt = txt + l

f = open(filename,"wb")
f.write(txt)
f.close()

if doneOk:
    print("swig_import_helper fixed for", filename)
else:
    print("Error: the swig import helper was not fixed, check", filename)
    print("       and fix this script: fix_swig_imports.py")
    exit(2)


exit(0)






#
# Example python script to generate a BOM from a KiCad generic netlist
#

"""
    @package
    Output: old OrcadPcb2 netlist format
    Sorted By: Ref

    Command line:
    python "pathToFile/netlist_form_OrcadPcb2.py" "%I" "%O.net"
"""

from __future__ import print_function

# Import the KiCad python helper module
import kicad_netlist_reader
import sys

# A helper function to convert a UTF8/Unicode/locale string read in netlist
# for python2 or python3 (Windows/unix)
def fromNetlistText( aText ):
    if sys.platform.startswith('win32'):
        try:
            return aText.encode('utf-8').decode('cp1252')
        except UnicodeDecodeError:
            return aText
    else:
        return aText

# Generate an instance of a generic netlist, and load the netlist tree from
# the command line option. If the file doesn't exist, execution will stop
netlist = kicad_netlist_reader.netlist(sys.argv[1])

# Open a file to write to, if the file cannot be opened output to stdout
# instead
canOpenFile = True
try:
    f = open(sys.argv[2], 'wb')
except IOError:
    e = "Can't open output file for writing: " + sys.argv[2]
    print(__file__, ":", e, sys.stderr)
    f = sys.stdout
    canOpenFile = False

components = netlist.getInterestingComponents( excludeBoard=True )

row =""

''' Netlist header:
( { EESchema Netlist Version 1.1  02/09/2021 11:10:00
Eeschema (5.99.0-12178-ge03257b55f-dirty)}
'''
''' Generate line ( { EESchema Netlist Version 1.1 <time> '''
row += '( { EESchema Netlist Version 1.1 ' + netlist.getDate() + '\n'
''' Generate line Eeschema <eeschema version> '''
row += 'Eeschema ' + netlist.getTool() + ' }\n\n'


'''
    This template read each component and creates lines:
     ( 3EBF7DBD $noname U1 74LS125
      ... pin list ...
      )
    and calls "create_pin_list" template to build the pin list
'''

for c in components:
    fp_name = c.getFootprint( False )

    if fp_name == "":
        fp_name = "$noname"

    row += "( " + c.getTimestamp() + ' ' + fp_name
    row += ' ' + c.getRef()
    row += ' ' + c.getValue()
    row += '\n'

    '''
    generate pin list
    The pin list from library description is something like
          <pins>
            <pin num="1" type="passive"/>
            <pin num="2" type="passive"/>
          </pins>
    Output pin list is ( <pin num> <net name> )
    something like
            ( 1 VCC )
            ( 2 GND )
    for not connected pins:
            ( 3 ? )
    '''
    lib = c.getLibPart()
    pinlist = lib.getPinList()

    for pin in pinlist:
        pin_num = pin.get('pin','num')
        row += "   ( " + pin_num + " "
        netname = c.getPinNetname( pin_num, netlist, True )

        if netname == "":
            netname = "?"

        row += netname
        row += " )\n"

    row += ')\n'

row += '\n)\n'

f.write(row.encode('utf-8'))
f.close()

if not canOpenFile:
    sys.exit(1)


#
# This Aogram source code file is part of KiCad, a free EDA CAD application.
#
# Copyright The KiCad Developers, see AUTHORS.txt for contributors.
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#

import shutil
from pathlib import Path
import pytest

from conftest import KiTestFixture
import utils


def get_fp_upgrade_cmd(input: str, output: str) -> list[str]:
    return [
        utils.kicad_cli(),
        "fp",
        "upgrade",
        "--force",
        str(input),
        "--output",
        str(output),
    ]


@pytest.mark.parametrize(
    "input_lib,output_dir,num_fps",
    [
        # This is a v9-formatted library with 2 footprints
        (
            "cli/fp_lib_test/Audio_Module.v9.pretty",
            "fp_lib_test",
            2,
        ),
    ],
)
def test_fp_upgrade(
    kitest: KiTestFixture,
    input_lib: str,
    output_dir: str,
    num_fps: int,
) -> None:
    """
    Perform an upgrade of a footprint library.
    """

    input_file = kitest.get_data_file_path(input_lib)
    output_path = kitest.get_output_path(f"cli/{output_dir}/upgrade/")

    shutil.rmtree(output_path, ignore_errors=True)
    output_path.mkdir(parents=True, exist_ok=True)


    upgraded_lib_path = Path(output_path) / (
        input_file.stem + ".upgraded.pretty"
    )

    upgrade_cmd = get_fp_upgrade_cmd(input_file, upgraded_lib_path)

    stdout, stderr, exitcode = utils.run_and_capture(upgrade_cmd)

    assert exitcode == 0
    assert stderr == ""
    assert stdout is not None

    assert upgraded_lib_path.is_dir()

    # Check the number of footprints in the upgraded library
    fp_files = list(upgraded_lib_path.glob("*.kicad_mod"))
    assert len(fp_files) == num_fps


@pytest.mark.parametrize(
    "input_lib,output_dir,num_fps",
    [
        # This is a v9-formatted library with 2 footprints
        (
            "cli/fp_lib_test/Audio_Module.v9.pretty",
            "fp_lib_test",
            2,
        ),
    ],
)
def test_fp_export_svg(
    kitest: KiTestFixture,
    input_lib: str,
    output_dir: str,
    num_fps: int,
) -> None:
    """
    Perform an export of a footprint library to SVG.

    Note that much of the functionality of the "fp export svg" command is tested by the
    PCB "pcb export svg" tests, since they share the same underlying code.
    """

    input_file = kitest.get_data_file_path(input_lib)
    output_path = kitest.get_output_path(f"cli/{output_dir}/svg/")

    output_path.mkdir(parents=True, exist_ok=True)

    export_cmd = [
        utils.kicad_cli(),
        "fp",
        "export",
        "svg",
        str(input_file),
        "--output",
        str(output_path),
    ]

    stdout, stderr, exitcode = utils.run_and_capture(export_cmd)

    assert exitcode == 0
    assert stderr == ""
    assert stdout is not None

    # Check the number of footprints in the exported directory
    fp_files = list(Path(output_path).glob("*.svg"))
    assert len(fp_files) == num_fps



#!/usr/bin/env python

# This program source code file is part of KiCad, a free EDA CAD application.
#
# Copyright (C) 2016 CERN
# @author Maciej Suminski <maciej.suminski@cern.ch>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, you may find one here:
# http://www.gnu.org/licenses/old-licenses/gpl-2.0.html
# or you may search the http://www.gnu.org website for the version 2 license,
# or you may write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA

# Converts a bitmap font atlas image in PNG format to raw pixel data stored
# in a C structure.

import os
import png
import sys

def convert(png_file):
    r = png.Reader(file=open(png_file, 'rb'))
    img = r.read()

    print(img);

    if img[3].get("alpha") == True:
        print('Detected alpha channel!');

    output = open(os.path.splitext(png_file)[0] + '_img.c', 'w')

    width = img[0]
    height = img[1]

    # Header
    output.write(
"""
/* generated with png2struct.py, do not modify by hand */

static const struct {
    unsigned int width, height;
    unsigned char pixels[%d * %d];
} bitmap_font = {
""" % (width, height));

    output.write('%d, %d,\n{' % (width, height))
    
    if img[3].get("alpha") == True:
        for row in img[2]:
            for p in row[3::4]:
                output.write('%d,' % p)
            output.write('\n');
    else:
        for row in img[2]:
            for p in row:
                output.write('%d,' % p)
            output.write('\n');

    output.write('}\n};\n')

    output.close()

#----------------------------------------------------------------------
if __name__ == "__main__":
    argc = len(sys.argv)

    if(argc == 2):
        convert(sys.argv[1])
    else:
        print("usage: %s <xml-file>" % sys.argv[0])



#
# Example python script to generate a BOM from a KiCad generic netlist
#

"""
    @package
    Output: PADS  format netlist
    Sorted By: Ref

    Command line:
    python "pathToFile/netlist_form_pads-pcb-asc.py" "%I" "%O.net"
"""

from __future__ import print_function

# Import the KiCad python helper module
import kicad_netlist_reader
import sys

# A helper function to convert a UTF8/Unicode/locale string read in netlist
# for python2 or python3 (Windows/unix)
def fromNetlistText( aText ):
    if sys.platform.startswith('win32'):
        try:
            return aText.encode('utf-8').decode('cp1252')
        except UnicodeDecodeError:
            return aText
    else:
        return aText

# Generate an instance of a generic netlist, and load the netlist tree from
# the command line option. If the file doesn't exist, execution will stop
netlist = kicad_netlist_reader.netlist(sys.argv[1])

# Open a file to write to, if the file cannot be opened output to stdout
# instead
canOpenFile = True
try:
    f = open(sys.argv[2], 'wb')
except IOError:
    e = "Can't open output file for writing: " + sys.argv[2]
    print(__file__, ":", e, sys.stderr)
    f = sys.stdout
    canOpenFile = False

components = netlist.getInterestingComponents( excludeBoard=True )

row =""

''' Netlist header '''
row += "!PADS-POWERPCB-V2.0-MILS!" + '\n'
row += '*PART*' + '\n'


''' Generate list of component
 for each component  create lines like
 C1 Capacitor_THT:CP_Axial_L18.0mm_D6.5mm_P25.00mm_Horizontal
 C2 Capacitor_THT:CP_Axial_L18.0mm_D6.5mm_P25.00mm_Horizontal
'''

for c in components:
    row += " " + c.getRef()

    fp_name = c.getFootprint( False )

    if fp_name != "":
        row += " " + fp_name

    row += '\n'

'''
generate for each net create lines like
*SIGNAL* /CLOCK-RB6
 P2.27
 P3.39
'''
nets = netlist.getNets()

row += '\n*NET*' + '\n'

for net in nets:
    # count the number of pads in net. nets with only one pad are skipped
    netitems = net.children
    pad_count = 0

    for node in netitems:
        pad_count += 1

    netitems = net.children

    if pad_count > 1:
        row += "*SIGNAL* " + net.get( "net", "name" ) + '\n'

        for node in netitems:
            row += node.get('node','ref') + '.' + node.get('node','pin') + '\n'

row += '*END*\n'

f.write(row.encode('utf-8'))
f.close()

if not canOpenFile:
    sys.exit(1)


#!/usr/bin/python

# Test a basic back to back FootprintLoad() of a single footprint, and FootprintEnumerate()

# 1) Build target _pcbnew after enabling scripting in cmake.
# $ make _pcbnew

# 2) Changed dir to pcbnew
# $ cd pcbnew
# $ pwd
# build/pcbnew

# 3) Entered following command line, script takes to arguments: library_path [footprint_name]
# $ PYTHONPATH=. <path_to>/test_plugin.py https://github.com/liftoff-sr/pretty_footprints   [100-LQFP]


from __future__ import print_function
from pcbnew import *
import sys

if len( sys.argv ) < 2 :
    print( "usage: script <library_path>  [<footprint_name>]" )
    sys.exit(1)


src_libpath = sys.argv[1]


src_type = IO_MGR.GuessPluginTypeFromLibPath( src_libpath );

src_plugin = IO_MGR.PluginFind( src_type )

if len( sys.argv ) == 2:
    list_of_footprints = src_plugin.FootprintEnumerate( src_libpath )
    for fp in list_of_footprints:
        print( fp )

elif len( sys.argv ) == 3:
    # I had some concerns about back to back reads, this verifies it is no problem:
    module = src_plugin.FootprintLoad( src_libpath, sys.argv[2] )
    if not module:
        print( "1st try: module", sys.argv[2], "not found" )
    module = src_plugin.FootprintLoad( src_libpath, sys.argv[2] )
    if not module:
        print( "2nd try: module", sys.argv[2], "not found" )
    print( module )





with open('CKJ_lib.lib') as fi, open('CKJ_wide.lib', 'w') as fo:
    for line in fi:
        tok = line.split()
        if tok:
            if tok[0] == 'P':
                tok[3:-2:2] = [str(int(t) * 1.5) for t in tok[3:-2:2]]
            elif tok[0] == 'X' and int(tok[3]) > 0:
                tok[3] = '1450'

        print(' '.join(tok), file=fo)



#
# This program source code file is part of KiCad, a free EDA CAD application.
#
# Copyright The KiCad Developers, see AUTHORS.txt for contributors.
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#

"""
Tests for kicad-cli pcb import command
"""

import utils
from pathlib import Path
import pytest
import json
import os
from conftest import KiTestFixture


def get_output_path( kitest: KiTestFixture, test_name: str, filename: str ) -> Path:
    """Generate an output path for a test file."""
    output_dir = kitest.get_output_path( f"cli/pcb_import/{test_name}/" )
    output_path = output_dir / filename

    if output_path.exists():
        output_path.unlink()

    return output_path


def get_pads_test_file() -> str:
    """Get path to a PADS test file from the QA test data directory."""
    test_data_dir = os.path.join( os.path.dirname( __file__ ), "..", "..", "data", "pcbnew", "plugins", "pads" )
    pads_file = os.path.join( test_data_dir, "ClaySight_MK1", "ClaySight_MK1.asc" )

    if os.path.exists( pads_file ):
        return pads_file

    return None


class TestPcbImportHelp:
    """Test help output for pcb import command"""

    def test_pcb_import_help( self, kitest: KiTestFixture ):
        """Test that pcb import --help shows usage information"""
        command = [ utils.kicad_cli(), "pcb", "import", "--help" ]

        stdout, _, return_code = utils.run_and_capture( command )

        assert return_code == 0
        assert "Import a non-KiCad PCB file" in stdout
        assert "--format" in stdout
        assert "--report-format" in stdout


class TestPcbImportErrors:
    """Test error handling for pcb import command"""

    def test_import_missing_file( self, kitest: KiTestFixture ):
        """Test error when input file doesn't exist"""
        output_path = get_output_path( kitest, "errors", "output.kicad_pcb" )

        command = [
            utils.kicad_cli(),
            "pcb", "import",
            "/nonexistent/path/file.asc",
            "-o", str( output_path )
        ]

        stdout, stderr, return_code = utils.run_and_capture( command )

        assert return_code != 0
        assert not output_path.exists()

    def test_import_invalid_format( self, kitest: KiTestFixture ):
        """Test error when invalid format specified"""
        output_path = get_output_path( kitest, "errors", "invalid_format.kicad_pcb" )

        command = [
            utils.kicad_cli(),
            "pcb", "import",
            "--format", "invalid_format",
            "/tmp/test.asc",
            "-o", str( output_path )
        ]

        stdout, stderr, return_code = utils.run_and_capture( command )

        assert return_code != 0


@pytest.mark.skipif( get_pads_test_file() is None,
                     reason="PADS test files not available at /tmp/JILA_pads_importer/" )
class TestPcbImportPads:
    """Test PADS file import"""

    def test_import_pads_file( self, kitest: KiTestFixture ):
        """Test importing a PADS file"""
        pads_file = get_pads_test_file()
        output_path = get_output_path( kitest, "pads", "imported.kicad_pcb" )

        command = [
            utils.kicad_cli(),
            "pcb", "import",
            "--format", "pads",
            pads_file,
            "-o", str( output_path )
        ]

        stdout, stderr, return_code = utils.run_and_capture( command )

        assert return_code == 0
        assert output_path.exists()
        assert output_path.stat().st_size > 0

        # Verify it's a valid KiCad file
        with open( output_path, "r" ) as f:
            content = f.read( 100 )
            assert "(kicad_pcb" in content

    def test_import_auto_detect( self, kitest: KiTestFixture ):
        """Test auto-detection of PADS format"""
        pads_file = get_pads_test_file()
        output_path = get_output_path( kitest, "pads", "auto_detected.kicad_pcb" )

        command = [
            utils.kicad_cli(),
            "pcb", "import",
            pads_file,
            "-o", str( output_path )
        ]

        stdout, stderr, return_code = utils.run_and_capture( command )

        assert return_code == 0
        assert output_path.exists()

    def test_import_json_report( self, kitest: KiTestFixture ):
        """Test JSON report generation"""
        pads_file = get_pads_test_file()
        output_path = get_output_path( kitest, "pads", "with_report.kicad_pcb" )
        report_path = get_output_path( kitest, "pads", "import_report.json" )

        command = [
            utils.kicad_cli(),
            "pcb", "import",
            "--format", "pads",
            "--report-format", "json",
            "--report-file", str( report_path ),
            pads_file,
            "-o", str( output_path )
        ]

        stdout, stderr, return_code = utils.run_and_capture( command )

        assert return_code == 0
        assert output_path.exists()
        assert report_path.exists()

        # Validate JSON report
        with open( report_path, "r" ) as f:
            report = json.load( f )

        assert "source_file" in report
        assert "source_format" in report
        assert "output_file" in report
        assert "statistics" in report
        assert "footprints" in report["statistics"]
        assert "tracks" in report["statistics"]
        assert "vias" in report["statistics"]
        assert "zones" in report["statistics"]

    def test_import_text_report( self, kitest: KiTestFixture ):
        """Test text report generation"""
        pads_file = get_pads_test_file()
        output_path = get_output_path( kitest, "pads", "with_text_report.kicad_pcb" )
        report_path = get_output_path( kitest, "pads", "import_report.txt" )

        command = [
            utils.kicad_cli(),
            "pcb", "import",
            "--format", "pads",
            "--report-format", "text",
            "--report-file", str( report_path ),
            pads_file,
            "-o", str( output_path )
        ]

        stdout, stderr, return_code = utils.run_and_capture( command )

        assert return_code == 0
        assert output_path.exists()
        assert report_path.exists()

        # Validate text report content
        with open( report_path, "r" ) as f:
            content = f.read()

        assert "Import Report" in content
        assert "Source file:" in content
        assert "Statistics:" in content
        assert "Footprints:" in content

    def test_import_json_report_stdout( self, kitest: KiTestFixture ):
        """Test JSON report output to stdout"""
        pads_file = get_pads_test_file()
        output_path = get_output_path( kitest, "pads", "report_stdout.kicad_pcb" )

        command = [
            utils.kicad_cli(),
            "pcb", "import",
            "--format", "pads",
            "--report-format", "json",
            pads_file,
            "-o", str( output_path )
        ]

        stdout, stderr, return_code = utils.run_and_capture( command )

        assert return_code == 0

        # Find JSON in stdout
        json_start = stdout.find( "{" )
        json_end = stdout.rfind( "}" ) + 1

        assert json_start >= 0
        assert json_end > json_start

        report = json.loads( stdout[json_start:json_end] )
        assert "statistics" in report



from __future__ import annotations

import pytest

from pybind11_tests import class_sh_virtual_py_cpp_mix as m


class PyBase(m.Base):  # Avoiding name PyDerived, for more systematic naming.
    def __init__(self):
        m.Base.__init__(self)

    def get(self):
        return 323


class PyCppDerived(m.CppDerived):
    def __init__(self):
        m.CppDerived.__init__(self)

    def get(self):
        return 434


@pytest.mark.parametrize(
    ("ctor", "expected"),
    [
        (m.Base, 101),
        (PyBase, 323),
        (m.CppDerivedPlain, 202),
        (m.CppDerived, 212),
        (PyCppDerived, 434),
    ],
)
def test_base_get(ctor, expected):
    obj = ctor()
    assert obj.get() == expected


@pytest.mark.parametrize(
    ("ctor", "expected"),
    [
        (m.Base, 4101),
        (PyBase, 4323),
        (m.CppDerivedPlain, 4202),
        (m.CppDerived, 4212),
        (PyCppDerived, 4434),
    ],
)
def test_get_from_cpp_plainc_ptr(ctor, expected):
    obj = ctor()
    assert m.get_from_cpp_plainc_ptr(obj) == expected


@pytest.mark.parametrize(
    ("ctor", "expected"),
    [
        (m.Base, 5101),
        (PyBase, 5323),
        (m.CppDerivedPlain, 5202),
        (m.CppDerived, 5212),
        (PyCppDerived, 5434),
    ],
)
def test_get_from_cpp_unique_ptr(ctor, expected):
    obj = ctor()
    assert m.get_from_cpp_unique_ptr(obj) == expected



#!/usr/bin/env -S uv run -q

# /// script
# dependencies = ["tomlkit"]
# ///
from __future__ import annotations

from pathlib import Path

import tomlkit

DIR = Path(__file__).parent.resolve()
PYPROJECT = DIR.parent / "pyproject.toml"


def get_global() -> str:
    pyproject = tomlkit.parse(PYPROJECT.read_text())
    del pyproject["tool"]["scikit-build"]["generate"]
    del pyproject["project"]["entry-points"]
    del pyproject["project"]["scripts"]
    del pyproject["tool"]["scikit-build"]["metadata"]["optional-dependencies"]
    pyproject["project"]["name"] = "pybind11-global"
    pyproject["tool"]["scikit-build"]["experimental"] = True
    pyproject["tool"]["scikit-build"]["wheel"]["install-dir"] = "/data"
    pyproject["tool"]["scikit-build"]["wheel"]["packages"] = []

    result = tomlkit.dumps(pyproject)
    assert isinstance(result, str)
    return result


if __name__ == "__main__":
    print(get_global())



#
# This program source code file is part of KiCad, a free EDA CAD application.
#
# Copyright (C) 2023 Mark Roszko <mark.roszko@gmail.com>
# Copyright (C) 2023 KiCad Developers
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#

import pytest
import os
from pathlib import Path

def pytest_report_header(config):
    """Print key environment vars."""
    path_val = os.environ.get("PATH", "")
    pythonpath_val = os.environ.get("PYTHONPATH", "")
    kibuildpaths_val = os.environ.get("KICAD_BUILD_PATHS", "")

    message = "Environment:\n"
    message += f"PATH={path_val}\n"
    message += f"PYTHONPATH={pythonpath_val}\n"
    message += f"KICAD_BUILD_PATHS={kibuildpaths_val}\n"

    return message


class KiTestFixture:
    junit: bool = False
    _output_path: Path = None
    _junit_folder: Path = None
    _ci_project_dir: Path = None

    def __init__( self, config ) -> None:
        self._junit = False

        env_project_dir = os.getenv( 'CI_PROJECT_DIR' )
        if env_project_dir is not None:
            self._ci_project_dir = Path( env_project_dir )

        junitxml = config.getoption("xmlpath")

        if junitxml is not None:
            p = Path( junitxml )
            p = Path( p.parent ) # get the directory as junitxml points to a file
            p = p.resolve() # get absolute path
            self._junit_folder = p
            self._junit = True
        else:
            p = Path.cwd()

        p = p.joinpath('output/')

        self._output_path = p

    def get_output_path( self, sub: str ) -> Path:
        """Return the calculated output path for test artifacts"""

        output_path =  self._output_path.joinpath( sub )

        os.makedirs( str( output_path ), exist_ok=True )

        return output_path

    def get_data_file_path( self, file: str ) -> Path:
        current_dir = os.path.dirname(__file__)
        base_data_path = os.path.abspath(os.path.join(current_dir, '../../data'))

        return Path(base_data_path).joinpath(file)

    def add_attachment( self, path: Path ) -> None:
        """Prints the attachment message line for junit reports"""

        if not self._junit:
            return

        # Make the attachment path relative, gitlab in particular wants it
        # relative tot he CI_PROJECT_DIR variable
        attach_src_path = path
        attach_path: Path = None
        if self._ci_project_dir is not None:
            attach_path = attach_src_path.relative_to( self._ci_project_dir )
        else:
            attach_path = attach_src_path.relative_to( self._junit_folder )

        print( "[[ATTACHMENT|{}]]".format( str( attach_path ) ) )


@pytest.fixture
def kitest( pytestconfig ):
    kitesthelper = KiTestFixture( pytestconfig )
    yield kitesthelper


# We need this on Windows when executing from ctest
if os.name == 'nt':
    for p in os.environ[ 'KICAD_BUILD_PATHS' ].split( ':' ):
        if os.path.isdir( p ):
            os.add_dll_directory( p )



import pytest
import pcbnew

class TestTracks:
    pcb : pcbnew.BOARD = None

    def setup_method(self):
        self.pcb = pcbnew.LoadBoard("../data/pcbnew/tracks_arcs_vias.kicad_pcb")

    def test_tracks(self):
        tracks = [t for t in self.pcb.Tracks() if t.GetClass() == 'PCB_TRACK']
        assert 16 == len(tracks)
        track = sorted(tracks, key=lambda t: [t.GetStart()[0], t.GetStart()[1]])[0]
        assert [27000000, 27585787] == [track.GetStart()[0], track.GetStart()[1]]
        assert [27000000, 26500000] == [track.GetEnd()[0], track.GetEnd()[1]]
        assert 250000 == track.GetWidth()
        assert 'McNetty' == track.GetNetname()

        dup_track = track.Duplicate()
        assert dup_track.m_Uuid != track.m_Uuid


    def test_arcs(self):
        arcs = [t.Cast() for t in self.pcb.Tracks() if t.GetClass() == 'PCB_ARC']
        assert 13 == len(arcs)
        arc = sorted(arcs, key=lambda t: [t.GetStart()[0], t.GetStart()[1]])[0]
        assert [29414200, 26500000] == [arc.GetCenter()[0], arc.GetCenter()[1]]
        assert [1800, 2250] == [round(arc.GetArcAngleStart().AsTenthsOfADegree()),
                                round(arc.GetArcAngleEnd().AsTenthsOfADegree())]
        assert 2414200 == round(arc.GetRadius())

        dup_arc = arc.Duplicate()
        assert dup_arc.m_Uuid != arc.m_Uuid 

    def test_vias(self):
        vias = [t.Cast() for t in self.pcb.Tracks() if t.GetClass() == 'PCB_VIA']
        assert 2 == len(vias)
        via = sorted(vias, key=lambda t: [t.GetStart()[0], t.GetStart()[1]])[0]
        assert [29000000, 41000000] == [via.GetStart()[0], via.GetStart()[1]]
        assert 400000 == via.GetDrillValue()

        dup_via = via.Duplicate()
        assert dup_via.m_Uuid != via.m_Uuid



#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#

from __future__ import division
import pcbnew

import FootprintWizardBase
import PadArray as PA


class RowedGridArray(PA.PadGridArray):

    def __init__(self, *args, **kwargs):
        super(RowedGridArray, self).__init__(*args, **kwargs)

    def NamingFunction(self, x, y):
        pad_cnt = self.nx*self.ny

        if self.ny == 1:
            return x+1

        if (y % 2) == 0:  # upper row, count down
            return pad_cnt-x
        else:  # lower row, count up
            return x+1


class RowedFootprint(FootprintWizardBase.FootprintWizard):

    pad_count_key = 'pad count'
    row_count_key = 'row count'
    row_spacing_key = 'row spacing'
    pad_length_key = 'pad length'
    pad_width_key = 'pad width'
    pad_pitch_key = 'pad pitch'

    silkscreen_inside_key = 'silk screen inside'
    outline_x_margin_key = 'outline x margin'
    outline_y_margin_key = 'outline y margin'

    def GenerateParameterList(self):
        # defaults for a DIP package
        self.AddParam("Pads", self.pad_count_key, self.uInteger, 24)
        self.AddParam("Pads", self.row_count_key, self.uInteger, 2, min_value=1, max_value=2)

        self.AddParam("Body", self.silkscreen_inside_key, self.uBool, False)
        self.AddParam("Body", self.outline_x_margin_key, self.uMM, 0.5)
        self.AddParam("Body", self.outline_y_margin_key, self.uMM, 0.5)

    def CheckParameters(self):
        self.CheckParam("Pads", self.pad_count_key, multiple=self.parameters['Pads'][self.row_count_key], info='Pads must be multiple of row count')

    def BuildThisFootprint(self):
        pads = self.parameters["Pads"]
        body = self.parameters["Body"]
        num_pads = pads[self.pad_count_key]
        pad_length = pads[self.pad_length_key]
        pad_width = pads[self.pad_width_key]
        row_pitch = pads[self.row_spacing_key]
        pad_pitch = pads[self.pad_pitch_key]
        num_rows = pads[self.row_count_key]

        pads_per_row = num_pads // num_rows

        # add in the pads
        pad = self.GetPad()

        array = RowedGridArray(pad, pads_per_row, num_rows, pad_pitch, row_pitch)
        array.AddPadsToModule(self.draw)

        # draw the Silk Screen
        Hsize = pad_pitch * (num_pads / num_rows - 1)
        Vsize = row_pitch * (num_rows - 1)
        pin1_posY = -Vsize / 2
        pin1_posX = -Hsize / 2

        pad_length = pads[self.pad_length_key]
        pad_width = pads[self.pad_width_key]

        ssx_offset = -pad_width / 2 - body[self.outline_x_margin_key]
        ssy_offset = -pad_length / 2 - body[self.outline_y_margin_key]

        if body[self.silkscreen_inside_key]:
            ssy_offset *= -1

        ssx = -pin1_posX - ssx_offset
        ssy = -pin1_posY - ssy_offset

        cmargin = self.draw.GetLineThickness()
        self.draw.SetLineThickness( pcbnew.FromMM( 0.12 ) ) #Default per KLC F5.1 as of 12/2018
        self.DrawBox(ssx, ssy)

        # Courtyard
        self.draw.SetLayer(pcbnew.F_CrtYd)
        cclearance = pcbnew.FromMM(0.25)
        sizex = (-pin1_posX + cclearance) * 2 + pad_width
        sizey = (-pin1_posY + cclearance) * 2 + pad_length
        # round size to nearest 0.02mm, rectangle will thus land on a 0.01mm grid
        sizex = pcbnew.PutOnGridMM(sizex, 0.02)
        sizey = pcbnew.PutOnGridMM(sizey, 0.02)
        # set courtyard line thickness to the one defined in KLC
        self.draw.SetLineThickness(pcbnew.FromMM(0.05))
        self.draw.Box(0, 0, sizex, sizey)
        # restore line thickness to previous value
        self.draw.SetLineThickness(pcbnew.FromMM(cmargin))

        #reference and value
        text_size = self.GetTextSize()  # IPC nominal

        if num_rows == 1:
            text_py = ssy + text_size
            self.draw.Value(0, -text_py, text_size)
            self.draw.Reference(0, text_py, text_size)
        else:
            text_px = ssx + text_size
            # self.draw.Value(text_px, 0, text_size, orientation_degree=90)
            self.draw.Value(0, 0, text_size)
            self.draw.Reference(-text_px, 0, text_size, orientation_degree=90)

        # set the attribute
        if self.GetName() == "S-DIP":
            self.module.SetAttributes(pcbnew.FP_THROUGH_HOLE)
        elif self.GetName() == "SOIC":
            self.module.SetAttributes(pcbnew.FP_SMD)

class SDIPWizard(RowedFootprint):

    def GetName(self):
        return "S-DIP"

    def GetDescription(self):
        return "Single/Dual Inline Package Footprint Wizard"

    def GenerateParameterList(self):
        RowedFootprint.GenerateParameterList(self)

        self.AddParam("Pads", self.pad_pitch_key, self.uMM, 2.54)
        self.AddParam("Pads", self.pad_width_key, self.uMM, 1.2)
        self.AddParam("Pads", self.pad_length_key, self.uMM, 2)
        self.AddParam("Pads", self.row_spacing_key, self.uMM, 7.52)
        self.AddParam("Pads", "drill size", self.uMM, 0.8)

    def GetValue(self):
        pads = self.parameters["Pads"]
        rows = pads[self.row_count_key]
        pad_count = pads[self.pad_count_key]
        row_dist_mil = pcbnew.pcbIUScale.IUToMils(int(self.parameters["Pads"][self.row_spacing_key])) #int(self.parameters["Pads"][self.row_spacing_key] / 2.54 * 100)
        pad_shape = ""

        if pads[self.pad_width_key] != pads[self.pad_length_key]:
            pad_shape = '_ELL'

        if rows == 1:
            name = "SIP"
            return "%s-%d" % (name, pad_count)

        name = "DIP"
        return "%s-%d_%d%s" % (name, pad_count, row_dist_mil, pad_shape)

    def GetPad(self):
        pad_length = self.parameters["Pads"][self.pad_length_key]
        pad_width = self.parameters["Pads"][self.pad_width_key]
        drill = self.parameters["Pads"]["drill size"]
        shape = pcbnew.PAD_SHAPE_CIRCLE

        if pad_length != pad_width:
            shape = pcbnew.PAD_SHAPE_OVAL

        return PA.PadMaker(self.module).THPad(
            pad_length, pad_width, drill, shape=shape)

    def DrawBox(self, ssx, ssy):

        if self.parameters["Pads"][self.row_count_key] == 2:

            #  ----------
            #  |8 7 6 5 |
            #  >        |
            #  |1 2 3 4 |
            #  ----------

            # draw the notch
            notchWidth = ssy/1.5
            notchHeight = self.draw.GetLineThickness()*3

            # NotchedBox draws the notch on top. Rotate the box 90 degrees
            # to have it on the left
            self.draw.NotchedBox(0, 0, ssy*2, ssx*2, notchWidth, notchHeight, -90)
        else:
            #  -----------------
            #  |1|2 3 4 5 6 7 8|
            #  -----------------
            self.draw.Box(0, 0, ssx*2, ssy*2)

            #line between pin1 and pin2
            pad_pitch = self.parameters["Pads"][self.pad_pitch_key]
            pad_cnt = self.parameters["Pads"][self.pad_count_key]
            line_x = ( pad_cnt/2 - 1) * pad_pitch
            self.draw.VLine(-line_x, -ssy, ssy * 2)

        return ssx, ssy

SDIPWizard().register()


class SOICWizard(RowedFootprint):

    def GetName(self):
        return "SOIC"

    def GetDescription(self):
        return "SOIC, MSOP, SSOP, TSSOP, etc, footprint wizard"

    def GetValue(self):
        pad_count = self.parameters["Pads"][self.pad_count_key]
        return "%s-%d" % ("SOIC", pad_count)

    def GenerateParameterList(self):
        RowedFootprint.GenerateParameterList(self)

        #and override some of them
        self.AddParam("Pads", self.pad_pitch_key, self.uMM, 1.27)
        self.AddParam("Pads", self.pad_width_key, self.uMM, 0.6)
        self.AddParam("Pads", self.pad_length_key, self.uMM, 2.2)
        self.AddParam("Pads", self.row_spacing_key, self.uMM, 5.2)

    def GetPad(self):
        pad_length = self.parameters["Pads"][self.pad_length_key]
        pad_width = self.parameters["Pads"][self.pad_width_key]
        return PA.PadMaker(self.module).SMDPad(
            pad_length, pad_width, shape=pcbnew.PAD_SHAPE_RECT)

    def DrawBox(self, ssx, ssy):

        #  ----------
        #  |8 7 6 5 |
        #  |1 2 3 4 |
        #  \---------

        setback = pcbnew.FromMM(0.8)

        if setback > ssy:
            setback = ssy

        self.draw.BoxWithDiagonalAtCorner(0, 0, ssx*2, ssy*2, setback, self.draw.flipY)

SOICWizard().register()



#
# This program source code file is part of KiCad, a free EDA CAD application.
#
# Copyright (C) 2023 Mark Roszko <mark.roszko@gmail.com>
# Copyright (C) 2023 Roberto Fernandez Bautista <roberto.fer.bau@gmail.com>
# Copyright (C) 2023 KiCad Developers
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#

import utils
from pathlib import Path
import pytest
import re
from typing import List, Tuple
from conftest import KiTestFixture
import sys


def get_generated_path(kitest: KiTestFixture,
                       input_file: Path,
                       test_name: str,
                       layer_name: str,
                       ) -> Tuple[Path, str]:
    layer_name_fixed = "-" + layer_name.replace( ".", "_" )
    generated_dir = str( kitest.get_output_path( "cli/{}/{}/".format( test_name, input_file.stem ) ) )
    generated_name = input_file.stem + layer_name_fixed + "-generated" + input_file.suffix
    generated_path = Path( generated_dir + "/" + generated_name )

    if generated_path.exists():
        generated_path.unlink()  # Delete file

    return [generated_path, layer_name_fixed]


def run_and_check_export_command(kitest: KiTestFixture,
                                 command: List[str],
                                 expected_output_file: Path):
    assert not expected_output_file.exists()

    stdout, stderr, exitcode = utils.run_and_capture( command )
    assert exitcode == 0
    # Don't assert stderr (legacy fills will have errors)
    assert stdout is not None
    assert expected_output_file.exists()

    kitest.add_attachment( expected_output_file )


@pytest.mark.parametrize("test_file,layers_to_test",
                         [
                            (
                                "cli/artwork_generation_regressions/ZoneFill-4.0.7.kicad_pcb",
                                ["F.Cu","B.Cu"]
                            ),
                            (   "cli/artwork_generation_regressions/ZoneFill-Legacy.brd",
                                ["F.Cu","B.Cu"]
                            )
                         ])
def test_pcb_export_svg( kitest: KiTestFixture,
                         test_file: str,
                         layers_to_test: List[str] ):

    input_file = kitest.get_data_file_path( test_file )

    for layer_name in layers_to_test:
        generated_svg_path, layer_name_fixed = get_generated_path( kitest,
                                                                   input_file.with_suffix( ".svg" ),
                                                                   "export_svg",
                                                                   layer_name )

        command = [utils.kicad_cli(), "pcb", "export", "svg", "--page-size-mode", "1",  # 1=Current page size
                   "--exclude-drawing-sheet", "--black-and-white", "--layers", layer_name,
                   "-o", str(generated_svg_path), str(input_file)]

        run_and_check_export_command( kitest, command, generated_svg_path )

        svg_source_path = str( input_file.with_suffix( "" ) )
        svg_source_path += layer_name_fixed + ".svg"

        # This test works only with Python >= 3.9 because it uses a pathlib function only existing
        # in 3.9 and newer. So skip it for previous versions
        if sys.hexversion >= 0x03090000 :
            # Comparison DPI = 1270 => 1px == 20um. I.e. allowable error of 60 um after eroding
            assert utils.svgs_are_equivalent( str( generated_svg_path ), svg_source_path, 1270,
                                              diff_handler=kitest.add_attachment )


@pytest.mark.skipif(not utils.is_gerbv_installed(), reason="Requires gerbv to be installed")
@pytest.mark.parametrize("test_file,layers_to_test,originInches,windowsizeInches",
                         [
                            (
                                "cli/artwork_generation_regressions/ZoneFill-4.0.7.kicad_pcb",
                                ["F.Cu","B.Cu"],
                                ( 3.5, -4.6 ),
                                ( 4.3, 2.2 )
                            ),
                            (   "cli/artwork_generation_regressions/ZoneFill-Legacy.brd",
                                ["F.Cu","B.Cu"],
                                ( -0.6, -0.4 ),
                                ( 1.7, 0.8 )
                            ),
                            (
                                # Regression for https://gitlab.com/kicad/code/kicad/-/issues/24143:
                                # gr_poly shapes whose outline contains a degenerate
                                # near-zero-width "spike" must keep that spike (drawn as a thick
                                # stroke) when exported to gerber. Fracture/Simplify drops
                                # the spike from the fill, so the stroke must be plotted from the
                                # unfractured outline.
                                "cli/artwork_generation_regressions/Issue24143.kicad_pcb",
                                ["F.Cu"],
                                ( 4.5, -3.1 ),
                                ( 0.4, 0.4 )
                            )
                         ])
def test_pcb_export_gerber( kitest: KiTestFixture,
                            test_file: str,
                            layers_to_test: List[str],
                            originInches :  Tuple[float, float],
                            windowsizeInches :  Tuple[float, float] ):

    input_file = kitest.get_data_file_path( test_file )

    for layer_name in layers_to_test:
        output_dir = kitest.get_output_path( "cli/export_gerber/{}/{}/".format(
                input_file.stem, layer_name.replace( ".", "_" ) ) )

        if output_dir.exists():
            import shutil
            shutil.rmtree( output_dir )

        output_dir.mkdir( parents=True, exist_ok=True )

        command = [utils.kicad_cli(), "pcb", "export", "gerbers", "--no-x2", "--use-drill-file-origin",
                   "--layers", layer_name,
                   "-o", str(output_dir), str(input_file)]

        stdout, stderr, exitcode = utils.run_and_capture( command )
        assert exitcode == 0
        assert stdout is not None

        gerber_files = [f for f in output_dir.iterdir() if not f.name.endswith( ".gbrjob" )]
        assert len( gerber_files ) == 1, \
            "Expected 1 gerber file in output dir, found {}".format( len( gerber_files ) )

        generated_gerber_path = gerber_files[0]
        kitest.add_attachment( generated_gerber_path )

        gbr_source_path = str( input_file.with_suffix( "" ) )
        layer_name_fixed = "-" + layer_name.replace( ".", "_" )
        gbr_source_path += layer_name_fixed + ".gbr"

        # Comparison DPI = 5080 => 1px == 5um. I.e. allowable error of 15 um after eroding
        assert utils.gerbers_are_equivalent( str( generated_gerber_path ), gbr_source_path, 5080,
                                             originInches, windowsizeInches,
                                             diff_handler=kitest.add_attachment )


@pytest.mark.parametrize("test_file,golden_name,output_dir,skip_line_count,cli_args",
                         [
                            (
                                "cli/basic_test/basic_test.kicad_pcb",
                                "basic_test_excellon_default.drl",
                                "basic_test/drills/excellon_default/",
                                5,
                                ["--format","excellon"]
                            ),
                            (
                                "cli/basic_test/basic_test.kicad_pcb",
                                "basic_test_excellon_inches.drl",
                                "basic_test/drills/excellon_inches/",
                                5,
                                ["--format","excellon","-u","in"]
                            ),
                            (
                                "cli/basic_test/basic_test.kicad_pcb",
                                "basic_test_excellon_mirror.drl",
                                "basic_test/drills/excellon_mirror/",
                                5,
                                ["--format","excellon","--excellon-mirror-y"]
                            ),
                            (
                                "cli/basic_test/basic_test.kicad_pcb",
                                "basic_test-PTH-drl.gbr",
                                "basic_test/drills/gerber_default/",
                                9,
                                ["--format","gerber"]
                            )
                         ])
def test_pcb_export_drill( kitest: KiTestFixture,
                         test_file: str,
                         golden_name: str,
                         output_dir: str,
                         skip_line_count: int,
                         cli_args: List[str]  ):

    input_file = kitest.get_data_file_path( test_file )

    output_path =  kitest.get_output_path( "cli/{}/".format( output_dir ) )

    command = [utils.kicad_cli(), "pcb", "export", "drill"]
    command.extend( cli_args )
    command.append( "-o" )
    command.append( str( output_path ) )
    command.append( str( input_file ) )

    stdout, stderr, exitcode = utils.run_and_capture( command )

    print(stdout)

    assert exitcode == 0
    assert stdout is not None

    stdout_regex = re.search("Created file '(.+)'", stdout)
    assert stdout_regex

    output_drill_path = Path( stdout_regex.group(1) )
    assert output_drill_path.exists()

    kitest.add_attachment( output_drill_path )

    compare_filepath = kitest.get_data_file_path( "cli/basic_test/{}".format( golden_name ) )
    assert utils.textdiff_files( compare_filepath, output_drill_path, skip_line_count )



#!/usr/bin/python

"""
Converts 5.1 font .lib files to .kicad_sym.
Deals with "rescaled" libraries that have non-integer coordinates,
such as CKJ_wide.lib.

This is not a general purpose converter - it's only meant to
deal with font files.

Usage: lib2sym.py
"""

import re
from dataclasses import dataclass, field
from typing import Any, ClassVar, TextIO

ROTATIONS = {'R': 0, 'U': 90, 'L': 180, 'D': 270}
VISIBILITIES = {'V': True, 'I': False}
ORIENTATIONS = {'H': 0, 'V': 90}
FILLTYPES = {'N': 'none'}
PINETYPES = {'I': 'input', 'O': 'output'}
PINSHAPES = {
    '': 'line',
    'I': 'inverted',
    'C': 'clock',
    'CI': 'inverted_clock',
    'L': 'input_low',
    'CL': 'clock_low',
    'F': 'falling_clock',
    'X': 'non_logic'}
PINHIDDEN = 'N'


def mil_to_mm(mil: float):
    return round(mil * 0.0254, 6)


def npairwise(iterable):
    args = (iter(iterable),) * 2
    return list(zip(*args))


class SexprWriter:
    def __init__(self, stream: TextIO):
        self.indent = 0
        self.stream = stream
        self.space = ""
        self.dedents: list[bool] = []

    def _indent(self) -> None:
        self.stream.write(" " * 2 * self.indent)

    def _newline(self) -> None:
        self.stream.write("\n")

    def _space(self) -> None:
        self.stream.write(self.space)
        self.space = ""

    def group(self, key: Any, /, items=None, newline=False, indent=False):
        self.startgroup(key, items, newline, indent)
        self.endgroup(newline=False)

    def startgroup(self, key: Any, /, items=None, newline=False, indent=False):
        self.dedents.append(indent)
        if indent:
            self.indent += 1
        if newline:
            self._newline()
            self._indent()
        else:
            self._space()
        self.stream.write("(")
        if key:
            self.stream.write(str(key))
            self.space = " "
        if items:
            for item in items:
                self.additem(item)

    def endgroup(self, newline=True):
        dedent = self.dedents.pop()
        if newline:
            self._newline()
            self._indent()
            self.space = ""
        else:
            self.space = " "
        if dedent:
            if self.indent > 0:
                self.indent -= 1
        self.stream.write(")")

    def additem(self, item: Any):
        self._space()
        if item == 0:
            self.stream.write("0")
        else:
            self.stream.write(str(item))
        self.space = " "


@dataclass
class Point:
    x: float
    y: float

    @classmethod
    def new_mil(cls, x: float, y: float) -> 'Point':
        return cls(mil_to_mm(x), mil_to_mm(y))

    def write(self, to: SexprWriter):
        to.group("xy", [self.x, self.y], newline=True, indent=True)


@dataclass
class TextEffect:
    size: float
    is_hidden: bool = False

    @classmethod
    def new_mil(cls, size: float) -> 'TextEffect':
        return cls(mil_to_mm(size))

    def write(self, to: SexprWriter, newline=True):
        if self:
            to.startgroup("effects", newline=newline, indent=True)
            to.startgroup("font")
            to.group("size", [self.size, self.size])
            to.endgroup(newline=False)
            if self.is_hidden:
                to.additem("hide")
            to.endgroup(newline=False)


@dataclass
class Field:
    name: str
    value: str
    posx: float
    posy: float
    rotation: float
    effects: TextEffect

    _pattern: ClassVar[re.Pattern] = re.compile(
        r'^\s*F(?P<n>\d+)\s+"(?P<value>[^"]*)"\s+(?P<rest>.+)$')

    @staticmethod
    def valid(line: str):
        return Field._pattern.match(line) is not None

    @classmethod
    def new_v5(cls, line: str) -> 'Field':
        match = Field._pattern.match(line)
        assert match is not None

        n = int(match.group("n"))
        name = ['Reference', 'Value', 'Footprint', 'Datasheet'][n]
        value = match.group("value")
        rest = match.group("rest").split()

        x = mil_to_mm(float(rest[0]))
        y = mil_to_mm(float(rest[1]))
        dimension = int(rest[2])
        rotation = ORIENTATIONS[rest[3]]
        visibility = VISIBILITIES[rest[4]]

        effects = TextEffect.new_mil(dimension)
        effects.is_hidden = not visibility
        return cls(name, value, x, y, rotation, effects)

    def write(self, to: SexprWriter):
        to.startgroup(
            "property", [
                f'"{self.name}"', f'"{self.value}"'], newline=True, indent=True)
        to.group("at", [self.posx, self.posy, self.rotation])
        self.effects.write(to)
        to.endgroup()


@dataclass
class Polyline:
    points: list[Point]
    stroke_width: float
    fill_type: str

    @staticmethod
    def valid(line: str):
        return line.startswith('P ')

    @classmethod
    def new_v5(cls, line: str) -> 'Polyline':
        tokens = line.split()

        points: list[Point] = []
        _1, _2, _3, stroke_width = tokens[1:5]
        fill_type = tokens[-1]

        for x, y in npairwise(tokens[5:]):
            points.append(Point.new_mil(float(x), float(y)))

        return cls(points,
                   mil_to_mm(int(stroke_width)),
                   FILLTYPES[fill_type]
                   )

    def write(self, to: SexprWriter):
        to.startgroup("polyline", newline=True, indent=True)

        to.startgroup("pts", newline=True, indent=True)
        for point in self.points:
            point.write(to)
        to.endgroup()

        to.startgroup("stroke", newline=True, indent=True)
        to.group("width", [self.stroke_width])
        to.group("type", ["solid"])
        to.endgroup(newline=False)

        to.startgroup("fill", newline=True, indent=True)
        to.group("type", [self.fill_type])
        to.endgroup(newline=False)

        to.endgroup()


@dataclass
class Pin:
    name: str
    number: str
    etype: str
    name_effect: TextEffect
    number_effect: TextEffect
    posx: float = 0.0
    posy: float = 0.0
    rotation: int = 0
    shape: str = "line"
    length: float = 2.54
    is_hidden: bool = False

    @staticmethod
    def valid(line: str):
        return line.startswith('X ')

    @classmethod
    def new_v5(cls, line: str) -> 'Pin':
        tokens = line.split()

        name, number, x, y, length, orientation, numsize, namesize, _1, _2, etype = tokens[
            1:12]
        shape = tokens[12] if len(tokens) > 12 else ''
        rotation = ROTATIONS[orientation]
        etype = PINETYPES[etype]
        if is_hidden := shape.startswith(PINHIDDEN):
            shape = shape[len(PINHIDDEN):]

        pin = cls(name, number, etype,
                  name_effect=TextEffect.new_mil(int(namesize)),
                  number_effect=TextEffect.new_mil(int(numsize))
                  )
        pin.posx = mil_to_mm(float(x))
        pin.posy = mil_to_mm(float(y))
        pin.shape = PINSHAPES[shape]
        pin.length = mil_to_mm(float(length))
        pin.is_hidden = is_hidden
        pin.rotation = rotation
        return pin

    def write(self, to: SexprWriter):
        to.startgroup("pin", [self.etype, self.shape],
                      newline=True, indent=True)

        to.group("at", [self.posx, self.posy, self.rotation])
        to.group("length", [self.length])

        to.startgroup("name", [f'"{self.name}"'], newline=True, indent=True)
        self.name_effect.write(to, newline=False)
        to.endgroup(newline=False)

        to.startgroup("number", [f'"{self.number}"'],
                      newline=True, indent=True)
        self.number_effect.write(to, newline=False)
        to.endgroup(newline=False)

        to.endgroup(newline=True)


@dataclass
class Symbol:
    name: str
    pin_names_offset: float
    properties: list[Field] = field(default_factory=list)
    pins: list[Pin] = field(default_factory=list)
    polylines: list[Polyline] = field(default_factory=list)

    @staticmethod
    def valid(line: str):
        return line.startswith("DEF ")

    @staticmethod
    def final(line: str):
        return line.startswith("ENDDEF")

    @classmethod
    def new_v5(cls, line: str) -> 'Symbol':
        tokens = line.split()
        name = tokens[1]
        offset = mil_to_mm(int(tokens[4]))
        return cls(name, offset)

    @property
    def has_contents(self):
        return bool(self.pins) | bool(self.polylines)

    def write(self, to: SexprWriter):
        to.startgroup("symbol", [f'"{self.name}"'], newline=True, indent=True)

        to.startgroup("pin_names")
        to.group("offset", [self.pin_names_offset])
        to.endgroup(newline=False)

        to.group("in_bom", ["yes"])
        to.group("on_board", ["yes"])

        for prop in self.properties:
            prop.write(to)

        if self.has_contents:
            to.startgroup(
                "symbol", [f'"{self.name}_0_1"'], newline=True, indent=True)
            for poly in self.polylines:
                poly.write(to)
            for pin in self.pins:
                pin.write(to)
            to.endgroup()

        to.endgroup()

    def getheft(self):
        """A good heuristic for how much space this symbol's sexprs take"""
        heft = 1 + 2*len(self.properties) + 3*len(self.pins)
        for poly in self.polylines:
            heft += 1 + len(poly.points)
        return heft


def translate(libfilename: str, symfilebase: str, symfileext: str, split=False):

    fo: Optional[TextIO] = None
    to: Optional[SexprWriter] = None
    part = ""

    def closeOutput():
        nonlocal fo, to
        if fo:
            to.endgroup()
            fo.write('\n')
            fo.close()
            fo = None

    def newOutput():
        nonlocal fo, to, part

        fo = open(f"{symfilebase}{part}{symfileext}", 'w', encoding="utf-8")
        to = SexprWriter(fo)
        to.startgroup("kicad_symbol_lib")
        to.group("version", ["20220914"])
        to.group("generator", ["font_lib2sym"])

    with open(libfilename, encoding="utf-8") as fi:

        linecount = 0
        totalheft = 0
        heftperfile = 175_000  # just shy of 10MB
        sym: Symbol

        while line := fi.readline():
            linecount += 1

            if Polyline.valid(line):
                polyline = Polyline.new_v5(line)
                sym.polylines.append(polyline)

            elif Pin.valid(line):
                pin = Pin.new_v5(line)
                sym.pins.append(pin)

            elif Field.valid(line):
                prop = Field.new_v5(line)
                sym.properties.append(prop)

            elif line.startswith('#'):
                pass

            elif Symbol.valid(line):
                sym = Symbol.new_v5(line)

            elif Symbol.final(line):
                if not fo or (split and totalheft > heftperfile):
                    if split:
                        part = f'_{sym.name}'
                    closeOutput()
                    newOutput()
                    totalheft = 0

                sym.write(to)
                totalheft += sym.getheft()
                if linecount & 127 == 1:
                    print(sym.name)

            elif line.startswith("DRAW") or line.startswith("ENDDRAW") or line.startswith("EESchema-LIBRARY"):
                pass

            elif line.strip():
                raise RuntimeError(f"Unknown line contents: {line}")

        closeOutput()


if __name__ == "__main__":
    fonts = ['symbol', 'hiragana', 'katakana',
             'half_full', 'font', 'CJK_symbol', 'CKJ_wide:split']

    for font in fonts:
        split = False
        out = font.replace("CKJ", "CJK")
        if font.endswith(":split"):
            font = font[:-6]
            out = out[:-6]
            split = True

        translate(f'{font}.lib', out, '.kicad_sym', split)



#!/usr/bin/env python3
"""
Development server for exercising the schematic remote symbol webview.

The server loads a JSON catalogue of parts, renders a small gallery UI, and
exposes buttons that stream each part's assets (footprint, models, etc.) to
KiCad via the WebView bridge. Each asset is compressed with zstd and then
base64-encoded before being embedded in the served HTML page.
"""

from __future__ import annotations

import argparse
import base64
import http.server
import hashlib
import json
import logging
import mimetypes
import socketserver
import urllib.parse
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Dict, List, Optional, Sequence

import zstandard
import re


_DEFAULT_PORT = 8080
_DEFAULT_CONFIG = (Path(__file__).parent / "data" / "parts.json").resolve()
_ZSTD_COMPRESSOR = zstandard.ZstdCompressor(level=10)


def _content_type_for_command(command: str) -> str:
  mapping = {
    "DL_SYMBOL": "KICAD_SYMBOL_V1",
    "DL_FOOTPRINT": "KICAD_FOOTPRINT_V1",
    "DL_SPICE": "KICAD_SPICE_MODEL_V1",
    "DL_3DMODEL": "KICAD_3D_MODEL_STEP",
  }

  return mapping.get(command.upper(), "UNKNOWN")


_GENERIC_IMAGE_DATA_URI = (
    "data:image/svg+xml;base64,"
    + base64.b64encode(
        b"""<svg xmlns='http://www.w3.org/2000/svg' width='160' height='100' viewBox='0 0 160 100'>
  <defs>
    <linearGradient id='g' x1='0%' y1='0%' x2='100%' y2='100%'>
      <stop offset='0%' stop-color='#e0e0e0'/>
      <stop offset='100%' stop-color='#bdbdbd'/>
    </linearGradient>
  </defs>
  <rect width='160' height='100' rx='12' fill='url(#g)'/>
  <path d='M25 75 L60 25 L100 25 L135 75 Z' fill='none' stroke='#616161' stroke-width='4' stroke-linecap='round' stroke-linejoin='round'/>
  <circle cx='80' cy='50' r='10' fill='none' stroke='#757575' stroke-width='4'/>
  <text x='80' y='90' font-family='Arial, sans-serif' font-size='18' fill='#424242' text-anchor='middle'>No Image</text>
</svg>"""
    ).decode("ascii")
)
_HTML_TEMPLATE = """\
<!DOCTYPE html>
<html lang="en">
  <head>
    <meta charset="utf-8" />
    <title>KiCad WebView Test Harness</title>
    <style>
      :root {{
        color-scheme: light;
        font-family: -apple-system, BlinkMacSystemFont, "Segoe UI", Arial, sans-serif;
        background: #e7eef8;
      }}

      body {{
        margin: 0;
        padding: 2rem;
        background: #e7eef8;
        color: #1f2933;
      }}

      .container {{
        max-width: 960px;
        margin: 0 auto;
        background: #fff;
        padding: 2rem;
        border-radius: 16px;
        box-shadow: 0 8px 30px rgba(15, 23, 42, 0.14);
      }}

      h1 {{
        margin-top: 0;
        margin-bottom: 0.5rem;
        font-size: 2rem;
      }}

      .intro {{
        margin-bottom: 2rem;
        color: #4b5563;
      }}

      .parts-grid {{
        display: grid;
        grid-template-columns: repeat(auto-fit, minmax(280px, 1fr));
        gap: 1rem;
      }}

      .login-panel {{
        display: flex;
        align-items: center;
        gap: 1rem;
        margin-bottom: 1.5rem;
      }}

      .login-status {{
        font-weight: 600;
        color: #1f2933;
      }}

      .part-card {{
        background: linear-gradient(135deg, #f9fafb, #f3f4f6);
        border-radius: 16px;
        padding: 1.5rem;
        box-shadow: inset 0 0 0 1px rgba(148, 163, 184, 0.25);
        display: flex;
        flex-direction: column;
        gap: 1rem;
      }}

      .part-card img {{
        width: 100%;
        height: 140px;
        object-fit: contain;
        border-radius: 12px;
        background: #fff;
        box-shadow: inset 0 0 0 1px rgba(99, 102, 241, 0.15);
      }}

      .part-name {{
        font-size: 1.2rem;
        margin: 0;
      }}

      .asset-list {{
        margin: 0;
        padding-left: 1.2rem;
        color: #4b5563;
        font-size: 0.95rem;
      }}

      button {{
        font-size: 1rem;
        padding: 0.75rem 1.25rem;
        border-radius: 999px;
        border: none;
        cursor: pointer;
        background: #2563eb;
        color: #fff;
        transition: background 0.15s ease, transform 0.15s ease;
      }}

      button:hover:enabled {{
        background: #1d4ed8;
        transform: translateY(-1px);
      }}

      button:disabled {{
        opacity: 0.6;
        cursor: not-allowed;
      }}

      pre {{
        background: #0f172a;
        color: #d1d5db;
        padding: 1rem;
        border-radius: 12px;
        overflow-x: auto;
        line-height: 1.5;
        min-height: 120px;
        font-size: 0.9rem;
      }}
    </style>
  </head>
  <body>
    <div class="container">
      <h1>Remote Symbol Demo Browser</h1>
      <p class="intro">
        Select a part to stream its assets to KiCad. Each asset is zstd-compressed,
        base64-encoded, and delivered in sequence with acknowledgement (ACK) handling.
      </p>
      <section class="login-panel">
        <button id="login-button" type="button">Log in via Browser</button>
        <span class="login-status" id="login-status">Not logged in</span>
      </section>
      <section class="parts-grid" id="parts-root"></section>
      <section>
        <h2>Transfer log</h2>
        <pre id="status-log">No transfers yet.</pre>
      </section>
    </div>
    <script>
      (function () {{
        const RPC_VERSION = 1;
        const PART_DEFINITIONS = Object.freeze({parts_json});
        const GENERIC_IMAGE = {generic_image};
        const STATUS_ELEMENT = document.getElementById("status-log");
        const PARTS_ROOT = document.getElementById("parts-root");
        const LOGIN_BUTTON = document.getElementById("login-button");
        const LOGIN_STATUS = document.getElementById("login-status");
        const RESPONSE_WAITERS = new Map();
        const BACKOFF_MS = [500, 1500, 3000];
        let messageCounter = 0;
        let sessionId = null;
        let loginInFlight = false;

        function postToKiCad(payload) {{
          if (window.webkit && window.webkit.messageHandlers && window.webkit.messageHandlers.kicad) {{
            window.webkit.messageHandlers.kicad.postMessage(payload);
            return true;
          }}

          if (window.chrome && window.chrome.webview && window.chrome.webview.postMessage) {{
            window.chrome.webview.postMessage(payload);
            return true;
          }}

          if (window.external && typeof window.external.invoke === "function") {{
            try {{
              window.external.invoke(payload);
              return true;
            }} catch (err) {{
              console.error("external.invoke failed", err);
            }}
          }}

          console.warn("No KiCad bridge found; simulated delivery only.");
          return false;
        }}

        function appendLog(message) {{
          const timestamp = new Date().toLocaleTimeString();
          const entry = `[${{timestamp}}] ${{message}}`;
          STATUS_ELEMENT.textContent =
            STATUS_ELEMENT.textContent === "No transfers yet."
              ? entry
              : `${{entry}}\\n${{STATUS_ELEMENT.textContent}}`;
        }}

        function updateLoginStatus(message) {{
          if (LOGIN_STATUS) {{
            LOGIN_STATUS.textContent = message;
          }}
        }}

        function generateFallbackUuid() {{
          return "xxxxxxxx-xxxx-4xxx-yxxx-xxxxxxxxxxxx".replace(/[xy]/g, (c) => {{
            const r = (Math.random() * 16) | 0;
            const v = c === "x" ? r : (r & 0x3) | 0x8;
            return v.toString(16);
          }});
        }}

        async function requestLogin(button) {{
          if (!sessionId) {{
            appendLog("Cannot log in until KiCad establishes a session.");
            return;
          }}

          if (loginInFlight) {{
            return;
          }}

          loginInFlight = true;
          if (button) {{
            button.disabled = true;
          }}

          updateLoginStatus("Requesting login status...");
          appendLog("Issuing REMOTE_LOGIN command...");

          try {{
            const response = await sendRpcCommand("REMOTE_LOGIN", {{ interactive: true }});
            const params = (response && response.parameters) ? response.parameters : {{}};

            if (params.authenticated === true) {{
              const label = params.provider_id ? ` (provider: ${{params.provider_id}})` : "";
              updateLoginStatus(`Authenticated${{label}}`);
              appendLog("Already authenticated.");
            }} else if (params.auth_type === "none") {{
              updateLoginStatus("Provider does not require authentication.");
              appendLog("Provider auth type is none.");
            }} else if (params.started === true) {{
              updateLoginStatus("Login flow started in external browser...");
              appendLog("OAuth login flow opened in system browser.");
            }} else {{
              updateLoginStatus("Login response received.");
              appendLog(`REMOTE_LOGIN: ${{JSON.stringify(params)}}`);
            }}
          }} catch (err) {{
            appendLog(`Remote login failed: ${{err.message}}`);
            updateLoginStatus("Login failed");
          }} finally {{
            loginInFlight = false;
            if (button) {{
              button.disabled = false;
            }}
          }}
        }}

        function installKiClientBridge() {{
          const existing = window.kiclient || {{}};
          const previousPost = typeof existing.postMessage === "function" ? existing.postMessage.bind(existing) : null;

          existing.postMessage = function (incoming) {{
            handleIncomingMessage(incoming);
            if (previousPost) {{
              previousPost(incoming);
            }}
          }};

          window.kiclient = existing;
        }}

        function handleIncomingMessage(incoming) {{
          let payload = incoming;
          if (typeof payload === "string") {{
            try {{
              payload = JSON.parse(payload);
            }} catch (err) {{
              console.warn("Failed to parse KiCad response", err, incoming);
              return;
            }}
          }}

          if (!payload || typeof payload !== "object") {{
            return;
          }}

          if (payload.command === "NEW_SESSION" && payload.response_to === undefined) {{
            sessionId = payload.session_id;
            messageCounter = payload.message_id || 0;
            RESPONSE_WAITERS.clear();
            appendLog(`Session established: ${{payload.session_id}}`);

            if (payload.user_id) {{
              updateLoginStatus(`Logged in as ${{payload.user_id}}`);
            }}

            sendNewSessionResponse(payload);
            return;
          }}

          if (payload.response_to !== undefined) {{
            const key = String(payload.response_to);
            const waiter = RESPONSE_WAITERS.get(key);
            if (waiter) {{
              RESPONSE_WAITERS.delete(key);
              waiter.resolve(payload);
            }} else {{
              appendLog(`Unexpected response for message ${{payload.response_to}}.`);
            }}
            return;
          }}

          appendLog(`KiCad -> ${{JSON.stringify(payload)}}`);
        }}

        function sendNewSessionResponse(request) {{
          if (!request || typeof request.message_id !== "number" || !sessionId) {{
            return;
          }}

          const msgId = ++messageCounter;
          const payload = {{
            version: RPC_VERSION,
            session_id: sessionId,
            message_id: msgId,
            response_to: request.message_id,
            command: "NEW_SESSION",
            status: "OK",
            parameters: {{
              server_name: "KiCad WebView Test Harness",
              server_version: "1.0.0"
            }}
          }};

          waitForResponse(msgId, 4000).catch(() => {{}});
          postToKiCad(JSON.stringify(payload));
        }}

        function waitForResponse(messageId, timeoutMs) {{
          return new Promise((resolve, reject) => {{
            const key = String(messageId);
            const timer = window.setTimeout(() => {{
              RESPONSE_WAITERS.delete(key);
              reject(new Error("Response timeout"));
            }}, timeoutMs);

            RESPONSE_WAITERS.set(key, {{
              resolve: (payload) => {{
                window.clearTimeout(timer);
                RESPONSE_WAITERS.delete(key);
                resolve(payload);
              }},
            }});
          }});
        }}

        function buildRpcEnvelope(command, parameters = {{}}, data = "") {{
          if (!sessionId) {{
            throw new Error("Session has not been established yet.");
          }}

          return {{
            version: RPC_VERSION,
            session_id: sessionId,
            message_id: ++messageCounter,
            command,
            parameters: JSON.parse(JSON.stringify(parameters)),
            data,
          }};
        }}

        async function sendRpcCommand(command, parameters = {{}}, data = "") {{
          const message = buildRpcEnvelope(command, parameters, data);
          const delivered = postToKiCad(JSON.stringify(message));

          if (!delivered) {{
            appendLog(`Could not deliver ${{command}} (message ${{message.message_id}}) to KiCad.`);
          }}

          const response = await waitForResponse(message.message_id, 4000);

          if (response.status && response.status !== "OK") {{
            const err = response.error_message || "Remote endpoint reported an error.";
            throw new Error(err);
          }}

          return response;
        }}

        async function sendAsset(asset) {{
          await sendRpcCommand(asset.command, asset.parameters, asset.data);
        }}

        async function sendWithRetries(asset) {{
          for (let attempt = 0; attempt < BACKOFF_MS.length; attempt += 1) {{
            try {{
              await sendAsset(asset);
              appendLog(`Transferred ${{asset.label}} (file: ${{asset.filename}})`);
              return;
            }} catch (err) {{
              appendLog(`Attempt ${{attempt + 1}} failed for ${{asset.label}}: ${{err.message}}`);

              if (attempt === BACKOFF_MS.length - 1) {{
                throw err;
              }}

              await new Promise((resolve) => window.setTimeout(resolve, BACKOFF_MS[attempt]));
            }}
          }}
        }}

        function formatBytes(bytes) {{
          if (!Number.isFinite(bytes)) {{
            return "";
          }}
          if (bytes < 1024) {{
            return `${{bytes}} B`;
          }}
          const kb = bytes / 1024;
          if (kb < 1024) {{
            return `${{kb.toFixed(1)}} KiB`;
          }}
          return `${{(kb / 1024).toFixed(2)}} MiB`;
        }}

        function renderParts() {{
          if (!PART_DEFINITIONS.length) {{
            PARTS_ROOT.innerHTML =
              '<p>No parts defined in the configuration file. Update data/parts.json to add entries.</p>';
            return;
          }}

          PART_DEFINITIONS.forEach((part) => {{
            const card = document.createElement("article");
            card.className = "part-card";

            const img = document.createElement("img");
            img.src = part.image || GENERIC_IMAGE;
            img.alt = `Preview for ${{part.name}}`;
            card.appendChild(img);

            const nameEl = document.createElement("h2");
            nameEl.className = "part-name";
            nameEl.textContent = part.name;
            card.appendChild(nameEl);

            const assetsList = document.createElement("ul");
            assetsList.className = "asset-list";

            part.assets.forEach((asset) => {{
              const item = document.createElement("li");
              const size = formatBytes(asset.size_bytes);
              item.textContent = size
                ? `${{asset.label}} · ${{asset.filename}} · ${{size}}`
                : `${{asset.label}} · ${{asset.filename}}`;
              assetsList.appendChild(item);
            }});

            card.appendChild(assetsList);

            const button = document.createElement("button");
            button.type = "button";
            button.textContent = `Send ${{part.name}} to KiCad`;
            button.addEventListener("click", () => handlePartTransfer(part, button));
            card.appendChild(button);

            PARTS_ROOT.appendChild(card);
          }});
        }}

        async function handlePartTransfer(part, button) {{
          if (!part.assets || !part.assets.length) {{
            appendLog(`No assets to transfer for ${{part.name}}.`);
            return;
          }}

          if (!sessionId) {{
            appendLog("Cannot transfer assets until KiCad establishes a session.");
            return;
          }}

          if (button.disabled) {{
            return;
          }}

          button.disabled = true;
          appendLog(`Beginning transfer for ${{part.name}}...`);

          try {{
            for (const asset of part.assets) {{
              await sendWithRetries(asset);
            }}
            appendLog(`Completed transfer for ${{part.name}}.`);
          }} catch (err) {{
            appendLog(`Transfer failed for ${{part.name}}: ${{err.message}}`);
          }} finally {{
            button.disabled = false;
          }}
        }}

        if (LOGIN_BUTTON) {{
          LOGIN_BUTTON.addEventListener("click", () => requestLogin(LOGIN_BUTTON));
        }}

        installKiClientBridge();
        renderParts();
      }})();
    </script>
  </body>
</html>
"""


_LOGIN_HTML_TEMPLATE = """\\
<!DOCTYPE html>
<html lang="en">
  <head>
    <meta charset="utf-8" />
    <title>KiCad Login</title>
    <style>
      :root {
        color-scheme: light;
        font-family: -apple-system, BlinkMacSystemFont, "Segoe UI", Arial, sans-serif;
        background: #e7eef8;
      }
      body {
        display: flex;
        justify-content: center;
        align-items: center;
        height: 100vh;
        margin: 0;
      }
      .login-card {
        background: #fff;
        padding: 2rem;
        border-radius: 16px;
        box-shadow: 0 8px 30px rgba(15, 23, 42, 0.14);
        width: 100%;
        max-width: 400px;
      }
      h1 { margin-top: 0; text-align: center; color: #1f2933; }
      .form-group { margin-bottom: 1rem; }
      label { display: block; margin-bottom: 0.5rem; font-weight: 600; color: #4b5563; }
      input {
        width: 100%;
        padding: 0.75rem;
        border: 1px solid #cbd5e1;
        border-radius: 8px;
        box-sizing: border-box;
        font-size: 1rem;
      }
      button {
        width: 100%;
        padding: 0.75rem;
        background: #2563eb;
        color: white;
        border: none;
        border-radius: 8px;
        font-size: 1rem;
        cursor: pointer;
        font-weight: 600;
        margin-top: 1rem;
      }
      button:hover { background: #1d4ed8; }
      #status { margin-top: 1rem; text-align: center; min-height: 1.5em; }
    </style>
  </head>
  <body>
    <div class="login-card">
      <h1>Sign In</h1>
      <form id="login-form">
        <div class="form-group">
          <label for="username">Username</label>
          <input type="text" id="username" name="username" required>
        </div>
        <div class="form-group">
          <label for="password">Password</label>
          <input type="password" id="password" name="password" required>
        </div>
        <button type="submit">Login</button>
      </form>
      <div id="status"></div>
    </div>
    <script>
      document.getElementById('login-form').addEventListener('submit', async (e) => {
        e.preventDefault();
        const params = new URLSearchParams(window.location.search);
        const port = params.get('port');
        const status = document.getElementById('status');

        if (!port) {
          status.textContent = "Error: No port provided in URL.";
          status.style.color = "#ef4444";
          return;
        }

        const username = document.getElementById('username').value;
        const uuid = (window.crypto && window.crypto.randomUUID) ? window.crypto.randomUUID() : "legacy-uuid-" + Math.random();

        // Construct the callback URL
        const callbackUrl = `http://localhost:${port}/login?user_id=${encodeURIComponent(uuid)}&username=${encodeURIComponent(username)}`;

        try {
          status.textContent = "Authenticating...";
          status.style.color = "#2563eb";

          await fetch(callbackUrl, { method: "GET", mode: "cors" });

          status.textContent = "Login successful! Redirecting...";
          status.style.color = "#16a34a";

          setTimeout(() => {
             window.location.href = "/";
          }, 1000);

        } catch (err) {
          console.error(err);
          status.textContent = "Login failed: " + err.message;
          status.style.color = "#ef4444";
        }
      });
    </script>
  </body>
</html>
"""


@dataclass(frozen=True)
class AssetRecord:
    """In-memory representation of a downloadable asset."""

    command: str
    label: str
    filename: str
    data: str
    parameters: Dict[str, Any]
    size_bytes: int

    def to_dict(self) -> Dict[str, Any]:
        return {
            "command": self.command,
            "label": self.label,
            "filename": self.filename,
            "data": self.data,
            "parameters": self.parameters,
            "size_bytes": self.size_bytes,
        }


@dataclass(frozen=True)
class PartRecord:
    """Renderable part entry."""

    name: str
    image: str
    assets: Sequence[AssetRecord]

    def to_dict(self) -> Dict[str, Any]:
        return {
            "name": self.name,
            "image": self.image,
            "assets": [asset.to_dict() for asset in self.assets],
        }


class _ServerState:
    """Holds parsed configuration and rendered HTML."""

    def __init__(self, config_path: Path) -> None:
        self.config_path = config_path
        self._parts = self._load_parts()
        self._page_bytes = self._render_page()
        self._login_page_bytes = _LOGIN_HTML_TEMPLATE.encode("utf-8")

    @property
    def page(self) -> bytes:
        return self._page_bytes

    @property
    def login_page(self) -> bytes:
        return self._login_page_bytes

    def build_provider_metadata(self, base_url: str) -> bytes:
        metadata = {
            "provider_name": "KiCad WebView Test Provider",
            "provider_version": "1.0.0",
            "api_base_url": base_url,
            "panel_url": f"{base_url}/",
            "auth": {"type": "none"},
            "capabilities": {
                "web_ui_v1": True,
                "inline_payloads_v1": True,
            },
            "max_download_bytes": 104857600,
            "supported_asset_types": ["symbol", "footprint", "3dmodel", "spice"],
            "allow_insecure_localhost": True,
        }
        return json.dumps(metadata).encode("utf-8")

    def _render_page(self) -> bytes:
        parts_payload = [part.to_dict() for part in self._parts]
        parts_json = json.dumps(parts_payload, separators=(",", ":")).replace("</", "<\\/")
        html = _HTML_TEMPLATE.format(parts_json=parts_json, generic_image=json.dumps(_GENERIC_IMAGE_DATA_URI))
        return html.encode("utf-8")

    def _load_parts(self) -> List[PartRecord]:
        if not self.config_path.is_file():
            raise FileNotFoundError(f"Config file '{self.config_path}' does not exist.")

        data = json.loads(self.config_path.read_text(encoding="utf-8"))
        if isinstance(data, dict):
            parts_data = data.get("Parts") or data.get("parts")
        elif isinstance(data, list):
            parts_data = data
        else:
            raise ValueError("Configuration must be a list or contain a 'Parts' array.")

        if not isinstance(parts_data, list):
            raise ValueError("Invalid configuration: expected an array of part objects.")

        parts: List[PartRecord] = []
        for index, entry in enumerate(parts_data, start=1):
            if not isinstance(entry, dict):
                raise ValueError(f"Part entry #{index} is not an object.")

            part = self._build_part(entry, index)
            parts.append(part)

        logging.info("Loaded %d part(s) from %s", len(parts), self.config_path)
        return parts

    def _build_part(self, entry: Dict[str, Any], index: int) -> PartRecord:
        name = self._get_field(entry, "Name")
        if not name:
            raise ValueError(f"Part #{index} is missing the 'Name' field.")

        symbol_path = self._get_field(entry, "Symbol")
        if not symbol_path:
            raise ValueError(f"Part '{name}' is missing the required 'Symbol' field.")

        assets: List[AssetRecord] = []
        component_entries: List[Dict[str, Any]] = []
        library_name = self._get_field(entry, "Library")
        if not library_name:
            library_name = Path(symbol_path).stem  # Fallback to symbol filename stem

        field_plan = [
          ("Footprint", "DL_FOOTPRINT", "SAVE", "footprint"),
          ("SPICE Model", "DL_SPICE", "SAVE", None),
          ("3D Model", "DL_3DMODEL", "SAVE", "3dmodel"),
          ("Symbol", "DL_SYMBOL", "PLACE", "symbol"),
        ]

        for label, command, mode, component_type in field_plan:
          path_value = self._get_field(entry, label)
          if not path_value:
            continue

          file_path = self._resolve_asset_path(path_value)
          raw_bytes = file_path.read_bytes()
          encoded = _encode_bytes(raw_bytes)
          extracted_name = _extract_asset_name(file_path, command)

          assets.append(
            AssetRecord(
              command=command,
              label=label,
              filename=file_path.name,
              data=encoded,
              parameters={
                "mode": mode,
                "compression": "ZSTD",
                "content_type": _content_type_for_command(command),
                "library": library_name,
                "name": extracted_name,
              },
              size_bytes=len(raw_bytes),
            )
          )

          if component_type:
            if component_type == "3dmodel":
              component_name = file_path.name
            else:
              component_name = extracted_name or file_path.stem

            component_entries.append(
              {
                "type": component_type,
                "name": component_name,
                "checksum": _hash_bytes(raw_bytes),
                "compression": "ZSTD",
                "content": encoded,
              }
            )

        if not assets:
          raise ValueError(f"Part '{name}' did not produce any downloadable assets.")

        if component_entries:
          component_json = json.dumps(component_entries, separators=(",", ":")).encode("utf-8")
          component_bundle = _encode_bytes(component_json)
          assets.insert(
            0,
            AssetRecord(
              command="DL_COMPONENT",
              label="Component Bundle",
              filename=f"{library_name}_component_bundle.json",
              data=component_bundle,
              parameters={
                "compression": "ZSTD",
                "library": library_name,
              },
              size_bytes=len(component_json),
            ),
          )

        image_path_value = self._get_field(entry, "Image")
        image_data = self._encode_image(image_path_value)

        return PartRecord(name=name, image=image_data, assets=assets)

    def _get_field(self, entry: Dict[str, Any], key: str) -> Optional[str]:
        direct = entry.get(key)
        if direct:
            return str(direct)

        lowered = key.lower()
        for candidate in (lowered, key.replace(" ", ""), key.replace(" ", "_").lower()):
            value = entry.get(candidate)
            if value:
                return str(value)

        return None

    def _resolve_asset_path(self, relative_path: str) -> Path:
        base = self.config_path.parent
        candidate = (base / relative_path).resolve()

        try:
            candidate.relative_to(base)
        except ValueError as err:
            raise ValueError(f"Asset path '{relative_path}' must stay within {base}") from err

        if not candidate.is_file():
            raise FileNotFoundError(f"Asset '{relative_path}' does not exist under {base}")

        return candidate

    def _encode_image(self, relative_path: Optional[str]) -> str:
        if not relative_path:
            return _GENERIC_IMAGE_DATA_URI

        image_path = self._resolve_asset_path(relative_path)
        mime = mimetypes.guess_type(image_path.name)[0] or "application/octet-stream"
        encoded = base64.b64encode(image_path.read_bytes()).decode("ascii")
        return f"data:{mime};base64,{encoded}"


def _encode_bytes(payload: bytes) -> str:
  compressed = _ZSTD_COMPRESSOR.compress(payload)
  return base64.b64encode(compressed).decode("ascii")


def _encode_file(file_path: Path) -> str:
  """Read, zstd-compress, and base64-encode a file."""

  return _encode_bytes(file_path.read_bytes())


def _hash_bytes(payload: bytes) -> str:
  return hashlib.sha256(payload).hexdigest()


def _extract_asset_name(file_path: Path, command: str) -> str:
  """Try to extract the internal symbol/footprint name from the file.

  Falls back to the filename stem when extraction fails or for binary
  assets.
  """
  try:
    text = file_path.read_text(encoding="utf-8", errors="ignore")
  except Exception:
    return file_path.stem

  cmd = (command or "").upper()

  # KiCad symbol libraries use s-expressions: (symbol "NAME" ...)
  if cmd == "DL_SYMBOL":
    m = re.search(r'\(symbol\s+"([^"]+)"', text)
    if m:
      return m.group(1)
    # Legacy library format (DEF ...)
    m = re.search(r'^\s*DEF\s+(\S+)', text, flags=re.MULTILINE)
    if m:
      return m.group(1)

  # Footprints: (footprint "NAME" ...) in modern .kicad_mod files
  if cmd == "DL_FOOTPRINT":
    m = re.search(r'\(footprint\s+"([^"]+)"', text)
    if m:
      return m.group(1)
    # Another possible older token is 'module <name>'
    m = re.search(r'^\s*module\s+([^\s(]+)', text, flags=re.MULTILINE)
    if m:
      return m.group(1)

  return file_path.stem


def _build_handler(state: _ServerState) -> type[http.server.BaseHTTPRequestHandler]:
    class _WebViewTestHandler(http.server.BaseHTTPRequestHandler):
        def do_GET(self) -> None:
            parsed = urllib.parse.urlparse(self.path)
            path = parsed.path

            if path in ("/", "/index.html"):
                content = state.page
                content_type = "text/html; charset=utf-8"
            elif path == "/login":
                content = state.login_page
                content_type = "text/html; charset=utf-8"
            elif path == "/.well-known/kicad-remote-provider":
                host = self.headers.get("Host") or f"localhost:{self.server.server_address[1]}"
                content = state.build_provider_metadata(f"http://{host}")
                content_type = "application/json; charset=utf-8"
            else:
                self.send_error(404)
                return

            self.send_response(200)
            self.send_header("Content-Type", content_type)
            self.send_header("Content-Length", str(len(content)))
            self.end_headers()
            self.wfile.write(content)

        def log_message(self, format: str, *args: Any) -> None:  # noqa: A003
            logging.info("%s - %s", self.client_address[0], format % args)

    return _WebViewTestHandler


def _parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser(description="Serve the KiCad WebView test page.")
    parser.add_argument(
        "--port",
        type=int,
        default=_DEFAULT_PORT,
        help=f"Local TCP port to bind (default: {_DEFAULT_PORT})",
    )
    parser.add_argument(
        "--config",
        type=Path,
        default=_DEFAULT_CONFIG,
        help=f"Path to the parts configuration JSON file (default: {_DEFAULT_CONFIG})",
    )

    return parser.parse_args()


def main() -> None:
    args = _parse_args()
    config_path = args.config.resolve()
    state = _ServerState(config_path=config_path)

    handler_cls = _build_handler(state)
    socketserver.TCPServer.allow_reuse_address = True

    with socketserver.TCPServer(("", args.port), handler_cls) as httpd:
        logging.info(
            "Serving WebView test page on http://localhost:%d/ using config %s",
            args.port,
            config_path,
        )
        logging.info("Press Ctrl+C to stop.")

        try:
            httpd.serve_forever()
        except KeyboardInterrupt:
            logging.info("Stopping server.")


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO, format="%(message)s")
    main()



#!/usr/bin/env python
import sys
from pcbnew import *

filename=sys.argv[1]

pcb = LoadBoard(filename)

for module in pcb.GetFootprints():
    if sys.version_info < (3, 0):
        print("* Module: %s" % module.GetReference().encode())
    else:
        print("* Module: %s" % module.GetReference())
    module.Value().SetVisible(False)      # set Value as Hidden
    module.Reference().SetVisible(True)   # set Reference as Visible

SaveBoard("mod_"+filename, pcb)



#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#

from __future__ import division
import pcbnew

import FootprintWizardBase

class FPC_FootprintWizard(FootprintWizardBase.FootprintWizard):

    def GetName(self):
        return "FPC (SMT connector)"

    def GetDescription(self):
        return "FPC (SMT connector) Footprint Wizard"

    def GetValue(self):
        pins = self.parameters["Pads"]["n"]
        return "FPC_%d" % pins

    def GenerateParameterList(self):
        self.AddParam( "Pads", "n", self.uInteger, 40 )
        self.AddParam( "Pads", "pitch", self.uMM, 0.5 )
        self.AddParam( "Pads", "width", self.uMM, 0.25 )
        self.AddParam( "Pads", "height", self.uMM, 1.6)
        self.AddParam( "Shield", "shield_to_pad", self.uMM, 1.6 )
        self.AddParam( "Shield", "from_top", self.uMM, 1.3 )
        self.AddParam( "Shield", "width", self.uMM, 1.5 )
        self.AddParam( "Shield", "height", self.uMM, 2 )


    # build a rectangular pad
    def smdRectPad(self,module,size,pos,name):
        pad = pcbnew.PAD(module)
        pad.SetSize(size)
        pad.SetShape(pcbnew.PAD_SHAPE_RECT)
        pad.SetAttribute(pcbnew.PAD_ATTRIB_SMD)
        pad.SetLayerSet( pad.SMDMask() )
        pad.SetPosition(pos)
        pad.SetName(name)
        return pad

    def CheckParameters(self):
        #TODO implement custom parameter checking
        pass

    def BuildThisFootprint(self):
        p = self.parameters
        pad_count       = int(p["Pads"]["n"])
        pad_width       = p["Pads"]["width"]
        pad_height      = p["Pads"]["height"]
        pad_pitch       = p["Pads"]["pitch"]
        shl_width       = p["Shield"]["width"]
        shl_height      = p["Shield"]["height"]
        shl_to_pad      = p["Shield"]["shield_to_pad"]
        shl_from_top    = p["Shield"]["from_top"]

        offsetX         = pad_pitch * ( pad_count-1 ) / 2
        size_pad = pcbnew.VECTOR2I( int(pad_width), int(pad_height) )
        size_shld = pcbnew.VECTOR2I(shl_width, shl_height)
        size_text = self.GetTextSize()  # IPC nominal

        # Gives a position and size to ref and value texts:
        textposy = pad_height/2 + pcbnew.FromMM(1) + self.GetTextThickness()
        angle_degree = 0.0
        self.draw.Reference( 0, textposy, size_text, angle_degree )

        textposy = textposy + size_text + self.GetTextThickness()
        self.draw.Value( 0, textposy, size_text )

        # create a pad array and add it to the module
        for n in range ( 0, pad_count ):
            xpos = (int)(pad_pitch*n - offsetX)
            pad = self.smdRectPad(self.module,size_pad, pcbnew.VECTOR2I( xpos, 0), str(n+1))
            self.module.Add(pad)


        # Mechanical shield pads: left pad and right pad
        xpos = -shl_to_pad-offsetX
        pad_s0_pos = pcbnew.VECTOR2I( int(xpos), int(shl_from_top) )
        pad_s0 = self.smdRectPad(self.module, size_shld, pad_s0_pos, "0")
        xpos = (pad_count-1) * pad_pitch+shl_to_pad - offsetX
        pad_s1_pos = pcbnew.VECTOR2I( int(xpos), int(shl_from_top) )
        pad_s1 = self.smdRectPad(self.module, size_shld, pad_s1_pos, "0")

        self.module.Add(pad_s0)
        self.module.Add(pad_s1)

        # add footprint outline
        self.draw.SetLineThickness( pcbnew.FromMM( 0.12 ) ) #Default per KLC F5.1 as of 12/2018
        linewidth = self.draw.GetLineThickness()
        margin = linewidth

        # upper line
        posy = -pad_height/2 - linewidth/2 - margin
        xstart = - pad_pitch*0.5-offsetX
        xend = pad_pitch * pad_count + xstart;
        min_y = posy
        max_y = posy
        self.draw.Line( xstart, posy, xend, posy )

        # lower line
        posy = pad_height/2 + linewidth/2 + margin
        max_y = max( max_y, posy )
        self.draw.Line(xstart, posy, xend, posy)

        # around left mechanical pad (the outline around right pad is mirrored/y axis)
        yend = pad_s0_pos.y + shl_height/2 + margin
        max_y = max( max_y, yend )
        self.draw.Line(xstart, posy, xstart, yend)
        self.draw.Line(-xstart, posy, -xstart, yend)

        posy = yend
        xend = pad_s0_pos.x - (shl_width/2 + linewidth + margin*2)
        max_x = -xend
        self.draw.Line(xstart, posy, xend, posy)

        # right pad side
        self.draw.Line(-xstart, posy, -xend, yend)

        # set SMD attribute
        self.module.SetAttributes(pcbnew.FP_SMD)

        # vertical segment at left of the pad
        xstart = xend
        yend = posy - (shl_height + linewidth + margin*2)
        self.draw.Line(xstart, posy, xend, yend)

        # right pad side
        self.draw.Line(-xstart, posy, -xend, yend)
        max_y = max( max_y, posy )
        min_y = min( min_y, yend )

        # horizontal segment above the pad
        xstart = xend
        xend = - pad_pitch*0.5-offsetX
        posy = yend
        self.draw.Line(xstart, posy, xend, yend)

        # right pad side
        self.draw.Line(-xstart, posy,-xend, yend)

        # vertical segment above the pad
        xstart = xend
        yend = -pad_height/2 - linewidth/2 - margin
        min_y = min( min_y, yend )
        self.draw.Line(xstart, posy, xend, yend)

        # right pad side
        self.draw.Line(-xstart, posy, -xend, yend)

        # Courtyard
        self.draw.SetLayer(pcbnew.F_CrtYd)
        max_x = max_x + linewidth + margin
        min_y = min_y - linewidth - margin
        max_y = max_y + linewidth + margin

        # round size to nearest 0.1mm, rectangle will thus land on a 0.05mm grid
        max_x = pcbnew.PutOnGridMM(max_x, 0.05)
        max_y = pcbnew.PutOnGridMM(max_y, 0.05)
        min_y = pcbnew.PutOnGridMM(min_y, 0.05)

        # set courtyard line thickness to the one defined in KLC
        self.draw.SetLineThickness(pcbnew.FromMM(0.05))
        self.draw.Line( -max_x, min_y, max_x, min_y )
        self.draw.Line( max_x, min_y, max_x, max_y )
        self.draw.Line( -max_x, max_y, max_x, max_y )
        self.draw.Line( -max_x, min_y, -max_x, max_y )

FPC_FootprintWizard().register()



#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#

from __future__ import division
import pcbnew

import FootprintWizardBase
import PadArray as PA


class RowedFootprint(FootprintWizardBase.FootprintWizard):

    def GenerateParameterList(self):

        # defaults for a ZIP package
        self.AddParam("Pads", "pad count", self.uInteger, 24)
        self.AddParam("Pads", "line count", self.uInteger, 2)

        self.AddParam("Body", "silkscreen inside", self.uBool, False)
        self.AddParam("Body", "courtyard margin", self.uMM, 0.5, min_value=0.2)

    @property
    def pads(self):
        return self.parameters['Pads']

    @property
    def body(self):
        return self.parameters['Body']

    def CheckParameters(self):
        # TODO - implement custom checks
        pass

    def BuildThisFootprint(self):

        pad_count = self.pads['pad count']
        pad_Vsize = self.pads['pad height']
        pad_Hsize = self.pads['pad width']
        line_pitch = self.pads['line spacing']
        pad_pitch = self.pads['pitch']
        line_count = self.pads['line count']

        if line_count == 1:
            singleline = True
        else:
            singleline = False

        # add in the pads
        pad = self.GetPad()

        array = PA.PadZGridArray(pad, pad_count, line_count, line_pitch, pad_pitch)
        array.AddPadsToModule(self.draw)

        # draw the Silk Screen
        pads_per_line = pad_count // line_count
        row_length = pad_pitch * (pads_per_line - 1)  # fenceposts
        ssx_offset = pad_Hsize / 2 + self.body['outline x margin']
        ssy_offset = pad_Vsize / 2 + self.body['outline y margin']

        pin1posX = pad_pitch * (pad_count - 1) / 2
        pin1posY = line_pitch * (line_count - 1) / 2
        leftx = pin1posX + ssx_offset
        lowy = pin1posY + ssy_offset

        cornery = lowy

        # body inside pads is possible only for 2 rows.
        # for other values, there is no room
        linew = self.draw.GetLineThickness()
        if self.body['silkscreen inside'] and line_count == 2:
            cornery = pin1posY - ssy_offset
            if cornery < linew:
                cornery = linew

        thick = self.draw.GetLineThickness()
        self.draw.SetLineThickness( pcbnew.FromMM( 0.12 ) ) #Default per KLC F5.1 as of 12/2018
        self.DrawBox(leftx*2, cornery*2)

        # Courtyard
        cmarginx = self.body['courtyard margin']
        cmarginy = cmarginx
        self.draw.SetLayer(pcbnew.F_CrtYd)
        sizex = (pin1posX + cmarginx) * 2 + pad_Hsize + thick
        sizey = (pin1posY + cmarginy) * 2 + pad_Vsize + thick
        # round size to nearest 0.1mm, rectangle will thus land on a 0.05mm grid
        sizex = pcbnew.PutOnGridMM(sizex, 0.1)
        sizey = pcbnew.PutOnGridMM(sizey, 0.1)
        # set courtyard line thickness to the one defined in KLC
        self.draw.SetLineThickness(pcbnew.FromMM(0.05))
        self.draw.Box(0, 0, sizex, sizey)
        # restore line thickness to previous value
        self.draw.SetLineThickness(pcbnew.FromMM(thick))

        #reference and value
        text_size = self.GetTextSize()  # IPC nominal
        t_posy = lowy + text_size

        self.draw.Value(0, t_posy, text_size)
        self.draw.Reference(0, -t_posy, text_size)

        # Add a extra text (${REFERENCE}) on the F_Fab layer
        extra_text = pcbnew.PCB_TEXT( self.module )
        extra_text.SetLayer( pcbnew.F_Fab )
        extra_text.SetPosition( pcbnew.VECTOR2I( 0, 0) )
        extra_text.SetTextSize( pcbnew.VECTOR2I( text_size, text_size ) )
        extra_text.SetText( "${REFERENCE}" )
        self.module.Add( extra_text )

        # set SMD attribute
        if self.GetName() == "ZIP":
            self.module.SetAttributes(pcbnew.FP_THROUGH_HOLE)
        elif self.GetName() == "ZOIC":
            self.module.SetAttributes(pcbnew.FP_SMD)

    def DrawBox(self, sizex, sizey):

        #  ----------
        #  | 2 4 6 8|
        #  |1 3 5 7 |
        #  \---------
        setback = pcbnew.FromMM(1)

        if setback > sizey/2:
            setback = sizey/2

        self.draw.BoxWithDiagonalAtCorner(0, 0, sizex, sizey, setback, self.draw.flipY)


class ZIPWizard(RowedFootprint):

    def GetName(self):
        return "ZIP"

    def GetDescription(self):
        return "N lines Zip Package Footprint Wizard"

    def GenerateParameterList(self):
        RowedFootprint.GenerateParameterList(self)

        self.AddParam("Pads", "pitch", self.uMM, 1.27)
        self.AddParam("Pads", "pad width", self.uMM, 1.2)
        self.AddParam("Pads", "pad height", self.uMM, 2)
        self.AddParam("Pads", "line spacing", self.uMM, 2.54)
        self.AddParam("Pads", "drill size", self.uMM, 0.8)
        self.AddParam("Body", 'outline x margin', self.uMM, 1)
        self.AddParam("Body", 'outline y margin', self.uMM, 0.5)

    def GetValue(self):
        rows = self.pads['line count']
        pad_cnt = self.pads['pad count']

        if rows == 1:
            name = "SIP"
        elif rows == 2:
            name = "ZIP"
        else:  # triple and up aren't really a thing, but call it something!
            name = "xIP"

        return "%s-%d" % (name, pad_cnt)

    def GetPad(self):
        pad_Vsize = self.pads['pad height']
        pad_Hsize = self.pads['pad width']
        drill = self.pads['drill size']
        return PA.PadMaker(self.module).THPad(
            pad_Vsize, pad_Hsize, drill, shape=pcbnew.PAD_SHAPE_OVAL)

ZIPWizard().register()


class ZOICWizard(RowedFootprint):

    def GetName(self):
        return "ZOIC"

    def GetDescription(self):
        return "ZOIC, etc, Footprint Wizard"

    def GetValue(self):
        return "%s-%d" % ("ZOIC-", self.pads['pad count'])

    def GenerateParameterList(self):
        RowedFootprint.GenerateParameterList(self)

        #and override some of them
        self.AddParam("Pads", "pitch", self.uMM, 0.6)
        self.AddParam("Pads", "pad width", self.uMM, 0.6)
        self.AddParam("Pads", "pad height", self.uMM, 1.8)
        self.AddParam("Pads", "line spacing", self.uMM, 5.2)

        self.AddParam("Body", "outline x margin", self.uMM, 0.5)
        self.AddParam("Body", "outline y margin", self.uMM, 1)

    def GetPad(self):
        pad_Vsize = self.pads['pad height']
        pad_Hsize = self.pads['pad width']
        return PA.PadMaker(self.module).SMDPad(
            pad_Vsize, pad_Hsize, shape=pcbnew.PAD_SHAPE_RECT)

ZOICWizard().register()



#!/usr/bin/env python3

"""
Bootstraps a dark-mode icon for any light-mode icon
"""

import argparse
import os
import re

# These were quickly chosen as a good starting point based on checking contrast against
# some light- and dark-mode window backgrounds taken from Ubuntu:
# light mode: #FCFCFC
# dark mode:  #484848

COLOR_MAP = {
#    "#ffffff": "#111111",  # white
#    "#fff":    "#111111",  # white (short)
    "#333333": "#E0E0E0",  # off black
    "#333":    "#E0E0E0",  # off black (short)
    "#545454": "#F4EFF3",  # dark grey primary
    "#606060": "#F0EBF0",  # dark grey large area
    "#909090": "#d0d0d0",  # medium grey
    "#b9b9b9": "#8f8f8f",  # light grey 1
    "#c1c1c1": "#999999",  # light grey 2
    "#f3f3f3": "#545454",  # off white 1
    "#f5f5f5": "#545454",  # off white 2
    "#1A81C4": "#42B8EB",  # primary blue
    "#39b4ea": "#1A81C4",  # light blue
    "#bf2641": "#f2647e",  # primary red
}


def process(src_dir, target_dir):
    for entry in os.scandir(src_dir):
        if entry.is_file() and entry.path.endswith(".svg"):
            processOne(os.path.abspath(entry.path), target_dir)

def processOne(src, target_dir):
    #print('Processing {}'.format(src))
    with open(src, 'r') as f:
        svg = f.read()

        for key in COLOR_MAP.keys():
            expr = re.compile(r'({})'.format(key), re.I)
            svg = expr.sub(COLOR_MAP[key], svg)

        target_name = os.path.join(target_dir, os.path.basename(src))

        with open(target_name, 'w') as out:
            out.write(svg)
            print('Wrote {}'.format(target_name))

if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument("source_dir")
    parser.add_argument("target_dir")
    args = parser.parse_args()

    process(args.source_dir, args.target_dir)



#!/usr/bin/env python

from __future__ import print_function
import sys
from pcbnew import *

# A helper function to convert a UTF8 string coming from Kicad for python2 or python3
def fromUTF8Text( aText ):
    if sys.version_info < (3, 0):
        return aText.encode()
    else:
        return aText

def GenerateBoard():
    size_0_6mm = VECTOR2I_MM(0.6,0.6)
    size_1_0mm = VECTOR2I_MM(1.0,1.0)

    # create a blank board
    pcb = CreateEmptyBoard()

    # create a new footprint, it's parent is our previously created pcb
    footprint = FOOTPRINT(pcb)
    footprint.SetReference("M1")   # give it a reference name
    footprint.Reference().SetPos(VECTOR2I_MM(3,-2))
    pcb.Add(footprint)             # add it to our pcb
    mod_pos = VECTOR2I_MM(50,50)
    footprint.SetPosition(mod_pos)

    # create a pad array and add it to the footprint
    n = 1
    for y in range (0,5):
        for x in range (0,5):
            pad = PAD(footprint)
            pad.SetDrillSize(size_0_6mm)
            pad.SetSize(size_1_0mm)
            pt = VECTOR2I_MM(1.27*x,1.27*y) + footprint.GetPosition()
            pad.SetPosition(pt);
            pad.SetName(str(n))
            footprint.Add(pad)
            n+=1


    # save the PCB to disk
    pcb.Save("my2.kicad_pcb")

def ReadBoard():
    pcb = LoadBoard("my2.kicad_pcb")

    for m in pcb.GetFootprints():
        print( 'footprint: ', fromUTF8Text( m.GetReference() ) )

        for p in m.Pads():
            print('pad ', fromUTF8Text( p.GetName() ), ToMM(p.GetPosition()), ToMM(p.GetOffset()))

GenerateBoard()
ReadBoard()



#!/usr/bin/env python

# Test the KiCad plugin regarding some expected features.

# 1) Build target _pcbnew after enabling scripting in cmake.
# $ make _pcbnew

# 2) Changed dir to pcbnew
# $ cd pcbnew
# $ pwd
# build/pcbnew

# 3) Entered following command line, script takes no arguments
# $ PYTHONPATH=. <path_to>/test_kicad_plugin.py

from __future__ import print_function
from pcbnew import IO_MGR, BOARD, FOOTPRINT, LIB_ID, UTF8
from os import rename as mv

tmp_path = '/tmp'
lib_path1 = "%s/lib1.pretty" % tmp_path
lib_path2 = "%s/lib2.pretty" % tmp_path

plugin = IO_MGR.PluginFind( IO_MGR.KICAD_SEXP )

# Expecting "KiCad":
print( "Plugin Type: %s" % plugin.PluginName() )

try:
    plugin.DeleteLibrary( lib_path1 )
except:
    pass    # ignore, new may not exist if first run

try:
    plugin.DeleteLibrary( lib_path2 )
except:
    pass    # ignore, new may not exist if first run

plugin.CreateLibrary( lib_path1 )

# Verify that the same plugin instance can edge trigger on a lib_path change
# for a CreateLibrary()
plugin.CreateLibrary( lib_path2 )

board = BOARD()

# The only way to construct a FOOTPRINT is to pass it a BOARD? Yep.
module = FOOTPRINT( board )

fpid = LIB_ID( '', 'mine' )

module.SetFPID( fpid )

plugin.FootprintSave( lib_path2, module )

# Verify that the same plugin instance can edge trigger on a lib_path change
# for a FootprintSave()
plugin.FootprintSave( lib_path1, module )

# create a disparity between the library's name ("footprint"),
# and the module's internal useless name ("mine").  Module is officially named "footprint" now
# but has (module mine ...) internally:
mv( "%s/mine.kicad_mod" % lib_path2, "%s/footprint.kicad_mod" % lib_path2 )

footprint = plugin.FootprintLoad( lib_path2, 'footprint' )

fpid = footprint.GetFPID()
fpid.SetLibNickname( UTF8( 'mylib' ) )
name = fpid.Format( fpid ).GetChars() # example to get the UTF8 char buffer

# Always after a FootprintLoad() the internal name should match the one used to load it.
print( "Internal name should be 'footprint': '%s'" % name  )

# Verify that the same plugin instance can edge trigger on a lib_path change
# for FootprintLoad()
footprint = plugin.FootprintLoad( lib_path1, 'mine' )

fpid = footprint.GetFPID()
fpid.SetLibNickname( UTF8( 'other_mylib' ) )

# Always after a FootprintLoad() the internal name should match the one used to load it.
# Example to print an UTF8 string
print( "Internal name should be 'mine': '%s'" % fpid.Format( fpid ) )

# As of 3-Dec-2013 this test is passed by KICAD_PLUGIN and Wayne is owed an atta boy!




#
# Example python script to generate an equivalent XML document from XML input
#
# Example: Round robin, XML to XML conversion
#

from __future__ import print_function

# Import the KiCad python helper module and the csv formatter
import kicad_netlist_reader
import sys
import pdb


# Generate an instance of a generic netlist, and load the netlist tree from
# the command line option. If the file doesn't exist, execution will stop
net = kicad_netlist_reader.netlist(sys.argv[1])

# Open a file to write to, if the file cannot be opened output to stdout
# instead
canOpenFile = True
try:
    f = open(sys.argv[2], 'w')
except IOError:
    e = "Can't open output file for writing: " + sys.argv[2]
    print( __file__, ":", e, sys.stderr)
    f = sys.stdout
    canOpenFile = False

print(net.formatXML(), file=f)

if not canOpenFile:
    sys.exit(1)


"""Unit-test parts of the CONNECTIVITY_DATA Python API."""

import pytest
import pcbnew


class TestConnectivity:
    """Test that calls on BOARD.Connectivity() are functional."""
    pcb : pcbnew.BOARD = None

    def setup_method(self):
        """Setup shared attributes."""
        self.pcb = pcbnew.LoadBoard("../data/pcbnew/complex_hierarchy.kicad_pcb")
        self.connectivity = self.pcb.GetConnectivity()

    def test_get_connectivity_returns_connectivity_data_object(self):
        """Verify: GetConnectivity() returns a CONNECTIVITY_DATA object."""
        connectivity = self.pcb.GetConnectivity()
        assert type(connectivity).__name__ == "CONNECTIVITY_DATA"

    def test_get_connected_pads_on_track_returns_iterable_of_pads(self):
        """Verify: GetConnectedPads(track) returns an iterable of pads."""
        tracks = list(self.pcb.TracksInNet(self.pcb.GetNetcodeFromNetname("/12Vext")))
        track_with_pad = tracks[1]
        """Note that this returns just the elements directly connected, not everything in the net"""
        pads = self.connectivity.GetConnectedPads(track_with_pad)
        assert len(pads) > 0
        assert all(pad.GetClass() == "PAD" for pad in pads)

    def test_get_connected_tracks_returns_iterable_of_tracks(self):
        """Verify: GetConnectedTracks(track) returns an iterable of tracks."""
        net_tracks = self.pcb.TracksInNet(self.pcb.GetNetcodeFromNetname("/12Vext"))
        net_track = list(net_tracks)[0]
        """Note that this returns just the elements directly connected, not everything in the net"""
        connected_tracks = self.connectivity.GetConnectedTracks(net_track)
        assert len(connected_tracks) > 1
        assert all(track.GetClass() == "PCB_TRACK" for track in connected_tracks)
